import os
from typing import List, Tuple

from jinja2 import Environment

from station.app.env import StationEnvironmentVariables
from station.common.constants import PHTDirectories, PHTImages, ServiceImages
from station.ctl.util import get_template_env


def render_compose(ctx: dict, env: Environment = None) -> str:
    """
    Render the docker-compose.yml file for the given config.

    Args:
        config: config dict
        env: template Environment

    Returns:

    """
    if not env:
        env = get_template_env()
    template = env.get_template("docker-compose.yml.tmpl")
    config = ctx["station_config"]

    service_images = {
        "db": ServiceImages.POSTGRES.value,
        "traefik": ServiceImages.TRAEFIK.value,
        "redis": ServiceImages.REDIS.value,
        "minio": ServiceImages.MINIO.value,
    }

    pht_images = {
        "airflow": PHTImages.AIRFLOW.value,
        "api": PHTImages.API.value,
        "ui": PHTImages.UI.value,
        "auth": PHTImages.AUTH.value,
    }
    host_path = config.get("host_path", None)

    install_dir = host_path if host_path else ctx["install_dir"]

    service_data_dir = os.path.join(install_dir, PHTDirectories.SERVICE_DATA_DIR.value)

    db_config = {
        "init_sql_path": ctx["init_sql_path"],
        "env": {
            "POSTGRES_USER": config["db"]["admin_user"],
            "POSTGRES_PASSWORD": config["db"]["admin_password"],
        },
    }

    certs_dir = str(os.path.join(install_dir, PHTDirectories.CERTS_DIR.value))

    proxy_config = {
        "http_port": config["http"]["port"],
        "https_port": config["https"]["port"],
        "labels": ["traefik.enable=true", "traefik.http.routers.traefik=true"],
        "traefik_config": ctx["traefik_config_path"],
        "router_config": ctx["router_config_path"],
        "certs_dir": certs_dir,
    }

    auth_url = "https://" + config["https"]["domain"] + "/auth"
    auth_config = {
        "env": {
            "ADMIN_PASSWORD": config["admin_password"],
            "NODE_ENV": "production",
            "PUBLIC_URL": auth_url,
            "AUTHORIZE_REDIRECT_URL": auth_url,
        },
        "port": config["auth"]["port"],
        "config_path": ctx["authup_config_path"],
        "labels": [
            "traefik.enable=true",
            "traefik.http.routers.auth.tls=true",
            f'traefik.http.routers.auth.rule=Host("{config["https"]["domain"]}") && PathPrefix("/auth")',
            "traefik.http.services.auth.loadbalancer.server.port=3010",
            "traefik.http.middlewares.auth-stripprefix.stripprefix.prefixes=/auth",
            "traefik.http.routers.auth.middlewares=auth-stripprefix@docker",
        ],
        "data_dir": str(os.path.join(service_data_dir, "auth")),
    }

    db_connection_string = (
        f"postgresql+psycopg2://{config['db']['admin_user']}:{config['db']['admin_password']}"
        f"@postgres/pht_station"
    )
    registry_user = config["registry"]["user"]
    if "$" in registry_user:
        registry_user = registry_user.replace("$", "$$")

    api_config = {
        "env": {
            "STATION_ID": config["id"],
            StationEnvironmentVariables.STATION_DB.value: db_connection_string,
            StationEnvironmentVariables.FERNET_KEY.value: config["api"]["fernet_key"],
            StationEnvironmentVariables.ENVIRONMENT.value: config["environment"],
            StationEnvironmentVariables.AIRFLOW_API_URL.value: f'https://{config["https"]["domain"]}/airflow/api/v1/',
            StationEnvironmentVariables.AIRFLOW_USER.value: config["airflow"][
                "admin_user"
            ],
            StationEnvironmentVariables.AIRFLOW_PW.value: config["airflow"][
                "admin_password"
            ],
            StationEnvironmentVariables.MINIO_HOST.value: "minio",
            StationEnvironmentVariables.MINIO_PORT.value: "9000",
            StationEnvironmentVariables.MINIO_ACCESS_KEY.value: config["minio"][
                "admin_user"
            ],
            StationEnvironmentVariables.MINIO_SECRET_KEY.value: config["minio"][
                "admin_password"
            ],
            StationEnvironmentVariables.REDIS_HOST.value: "redis",
            StationEnvironmentVariables.AUTH_SERVER_HOST.value: "http://auth",
            StationEnvironmentVariables.AUTH_SERVER_PORT.value: config["auth"]["port"],
            StationEnvironmentVariables.AUTH_ROBOT_ID.value: config["auth"]["robot_id"],
            StationEnvironmentVariables.AUTH_ROBOT_SECRET.value: config["auth"][
                "robot_secret"
            ],
            StationEnvironmentVariables.REGISTRY_URL.value: config["registry"][
                "address"
            ],
            StationEnvironmentVariables.REGISTRY_USER.value: registry_user,
            StationEnvironmentVariables.REGISTRY_PW.value: config["registry"][
                "password"
            ],
            StationEnvironmentVariables.REGISTRY_PROJECT.value: config["registry"][
                "project"
            ],
            StationEnvironmentVariables.CENTRAL_API_URL.value: config["central"]["url"],
            StationEnvironmentVariables.STATION_ROBOT_ID.value: config["central"][
                "robot_id"
            ],
            StationEnvironmentVariables.STATION_ROBOT_SECRET.value: config["central"][
                "robot_secret"
            ],
            StationEnvironmentVariables.STATION_DATA_DIR.value: config["data_dir"],
        },
        "labels": [
            "traefik.enable=true",
            "traefik.http.routers.api.tls=true",
            f'traefik.http.routers.api.rule=Host("{config["https"]["domain"]}") && PathPrefix("/api")',
            "traefik.http.services.api.loadbalancer.server.port=8000",
        ],
    }

    minio_config = {
        "env": {
            "MINIO_ROOT_USER": config["minio"]["admin_user"],
            "MINIO_ROOT_PASSWORD": config["minio"]["admin_password"],
        },
        "labels": [
            "traefik.enable=true",
            "traefik.http.routers.minio.tls=true",
            f'traefik.http.routers.minio.rule=Host("minio.{config["https"]["domain"]}")',
            "traefik.http.routers.minio.service=minio",
            "traefik.http.services.minio.loadbalancer.server.port=9000",
            "traefik.http.routers.minio-console.tls=true",
            f'traefik.http.routers.minio-console.rule=Host("minio-console.{config["https"]["domain"]}")',
            "traefik.http.routers.minio-console.service=minio-console",
            "traefik.http.services.minio-console.loadbalancer.server.port=9001",
        ],
    }

    if config.get("host_path"):
        key_name = config["central"]["private_key"].split("/")[-1]
        key_path = os.path.join(config["host_path"], key_name)
    else:
        key_path = config["central"]["private_key"]
    airflow_config = {
        "private_key": key_path,
        "config_path": ctx["airflow_config_path"],
        "env": {
            "STATION_ID": config["registry"]["project"],
            "AIRFLOW_USER": config["airflow"]["admin_user"],
            "AIRFLOW_PW": config["airflow"]["admin_password"],
            "HARBOR_URL": config["registry"]["address"],
            "HARBOR_USER": registry_user,
            "HARBOR_PW": config["registry"]["password"],
        },
        "labels": [
            "traefik.enable=true",
            "traefik.http.routers.airflow.tls=true",
            f'traefik.http.routers.airflow.rule=Host("{config["https"]["domain"]}") && PathPrefix("/airflow")',
            "traefik.http.services.airflow.loadbalancer.server.port=8080",
        ],
    }

    station_data_dir = str(
        os.path.join(ctx["install_dir"], PHTDirectories.STATION_DATA_DIR.value)
    )

    ui_config = {
        "env": {
            "STATION_API_URL": "https://" + config["https"]["domain"] + "/api",
            "AUTH_API_URL": "https://" + config["https"]["domain"] + "/auth/",
        },
        "labels": [
            "traefik.enable=true",
            "traefik.http.routers.ui.tls=true",
            f'traefik.http.routers.ui.rule=Host("{config["https"]["domain"]}") && !PathPrefix("/api") '
            f'&& !PathPrefix("/auth") && !PathPrefix("/airflow") && !PathPrefix("/fhir-servers")',
            "traefik.http.services.ui.loadbalancer.server.port=3000",
        ],
    }

    return template.render(
        service_images=service_images,
        pht_images=pht_images,
        version=config["version"],
        service_data_dir=service_data_dir,
        station_data_dir=station_data_dir,
        db_config=db_config,
        proxy_config=proxy_config,
        auth_config=auth_config,
        api_config=api_config,
        minio_config=minio_config,
        airflow_config=airflow_config,
        ui_config=ui_config,
    )


def render_airflow_config(
    domain: str, sql_alchemy_conn: str, env: Environment = None
) -> str:
    if not env:
        env = get_template_env()

    template = env.get_template("airflow.cfg.tmpl")
    airflow_base_url = "https://" + domain + "/airflow"
    return template.render(
        airflow_base_url=airflow_base_url, sql_alchemy_conn=sql_alchemy_conn
    )


def render_authup_api_config(
    auth_config: dict,
    db_user: str,
    db_password: str,
    env: Environment | None = None,
) -> str:
    if not env:
        env = get_template_env()

    template = env.get_template("authup/authup.api.conf.tmpl")

    return template.render(
        auth_config=auth_config, db_user=db_user, db_password=db_password
    )


def render_traefik_configs(
    http_port: int = 80,
    https_port: int = 443,
    https_enabled: bool = True,
    domain: str = None,
    certs: dict = None,
    env: Environment = None,
) -> Tuple[str, str]:
    """
    Render static config files for the traefik proxy.

    Args:
        http_port: which port to use for http traffic
        https_port: which port to use for https traffic
        https_enabled: boolean whether to enable https traffic
        domain: domain to use for https traffic
        certs: certificates for the given domain
        env: template Environment

    Returns: Tuple of the traefik config and router config yaml files as strings

    """

    # initialize environment if it is not given
    if not env:
        env = get_template_env()

    # render traefik config
    traefik_config = _make_traefik_config(
        env=env,
        http_port=http_port,
        https_port=https_port,
        https_enabled=https_enabled,
        dashboard=True,
        domain=domain,
    )

    certs = [
        {
            "cert": f"/etc/certs/{certs['cert'].split('/')[-1]}",
            "key": f"/etc/certs/{certs['key'].split('/')[-1]}",
        }
    ]

    # render traefik router config
    router_config = _make_traefik_router_config(
        env=env, https_enabled=https_enabled, domain=domain, certs=certs
    )

    return traefik_config, router_config


def render_init_sql(db_user: str, env: Environment = None) -> str:
    """
    Render the init.sql file for setting up the postgres database.
    The given user and two databases will be created with this script, the user is given full permissions on all
    created databases.

    Args:
        db_user: username for the DBMS
        env: template Environment

    Returns:

    """
    if not env:
        env = get_template_env()
    template = env.get_template("init.sql.tmpl")
    return template.render(db_user=db_user)


def _make_traefik_config(
    env: Environment,
    http_port: int = 80,
    https_port: int = None,
    https_enabled: bool = True,
    domain: str = None,
    dashboard: bool = False,
) -> str:
    """
    Render the general traefik config file.

    Args:
        env: template Environment
        http_port: port to use for http traffic
        https_port: port to use for https traffic
        https_enabled: https enabled
        dashboard: whether to enable the traefik dashboard

    Returns: string containing the content of the traefik config yaml file

    """
    template = env.get_template("traefik/traefik.yml.tmpl")
    return template.render(
        dashboard=dashboard,
        http_port=http_port,
        https_port=https_port,
        https_enabled=https_enabled,
        domain=domain,
    )


def _make_traefik_router_config(
    env: Environment,
    https_enabled: bool = True,
    domain: str = None,
    certs: List[dict] = None,
) -> str:
    """
    Render the traefik router config file. This file contains static router configuration for the traefik proxy.
    As well the specifications on which domains and respective certificates to use for https traffic.
    Args:
        env: template Environment
        https_enabled: whether to enable https traffic
        domain: domain to use for https traffic
        certs: certificates for the given domain

    Returns: string containing the content of the traefik router config yaml file

    """
    template = env.get_template("traefik/config.yml.tmpl")
    return template.render(domain=domain, https_enabled=https_enabled, certs=certs)



from builtins import str
from datetime import datetime
from typing import List, Tuple, Union

from dateutil import parser
from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from station.app.config import clients, settings
from station.app.models.docker_trains import (
    DockerTrain,
    DockerTrainConfig,
    DockerTrainExecution,
    DockerTrainState,
)
from station.app.schemas.docker_trains import (
    DockerTrainConfigCreate,
    DockerTrainCreate,
    DockerTrainUpdate,
)
from station.app.schemas.docker_trains import DockerTrainState as DockerTrainStateSchema

# TODO improve handling of proposals
from .base import CRUDBase, ModelType


class CRUDDockerTrain(CRUDBase[DockerTrain, DockerTrainCreate, DockerTrainUpdate]):
    def create(self, db: Session, *, obj_in: DockerTrainCreate) -> ModelType:

        if isinstance(obj_in.config, int):
            db_config = (
                db.query(DockerTrainConfig)
                .filter(DockerTrainConfig.id == obj_in.config)
                .first()
            )
            if not db_config:
                raise HTTPException(
                    status_code=404, detail=f"Config {obj_in.config} not found"
                )
            config_id = db_config.id

        elif isinstance(obj_in.config, DockerTrainConfigCreate):
            db_config: DockerTrainConfig = (
                db.query(DockerTrainConfig)
                .filter(DockerTrainConfig.name == obj_in.config.name)
                .first()
            )
            if db_config:
                raise HTTPException(
                    status_code=400,
                    detail="A config with the given name already exists.",
                )
            else:
                new_config = DockerTrainConfig(**jsonable_encoder(obj_in.config))
                db.add(new_config)
                db.commit()
                db.refresh(new_config)
                config_id = new_config.id

        else:
            config_id = None

        db_train = DockerTrain(train_id=obj_in.train_id, config_id=config_id)
        db.add(db_train)
        db.commit()
        db.refresh(db_train)
        train_state = DockerTrainState(train_id=db_train.id)
        db.add(train_state)
        db.commit()

        db.refresh(db_train)
        return db_train

    def get_by_train_id(self, db: Session, train_id: str) -> DockerTrain:
        train = db.query(DockerTrain).filter(DockerTrain.train_id == train_id).first()
        return train

    def get_trains_by_active_status(
        self, db: Session, active=True, limit: int = 0
    ) -> List[DockerTrain]:
        if limit != 0:
            trains = (
                db.query(DockerTrain)
                .filter(DockerTrain.is_active == active)
                .limit(limit)
                .all()
            )
        else:
            trains = db.query(DockerTrain).filter(DockerTrain.is_active == active).all()
        return trains

    def delete_by_train_id(self, db: Session, train_id: str) -> DockerTrain:
        db_train = self.get_by_train_id(db, train_id)
        if not db_train:
            raise HTTPException(status_code=404, detail=f"Train {train_id} not found")
        db.delete(db_train)
        db.commit()
        return db_train

    def add_if_not_exists(
        self,
        db: Session,
        train_id: str,
        created_at: str = datetime.now(),
        updated_at: str = None,
    ):
        db_train = self.get_by_train_id(db, train_id)
        if not db_train:
            if updated_at:
                db_train = DockerTrain(
                    train_id=train_id,
                    created_at=parser.parse(created_at),
                    updated_at=parser.parse(updated_at),
                )
            else:
                db_train = DockerTrain(
                    train_id=train_id, created_at=parser.parse(created_at)
                )
            db.add(db_train)
            db.commit()
            db.refresh(db_train)
            train_state = DockerTrainState(train_id=db_train.id)
            db.add(train_state)
            db.commit()
            return db_train

    def read_train_state(self, db: Session, train_id: str) -> DockerTrainState:
        db_train = self.get_by_train_id(db, train_id)
        if not db_train:
            raise HTTPException(status_code=404, detail=f"Train {train_id} not found")
        state = db_train.state
        return state

    def update_train_state(
        self, db: Session, train_id: str, state_in: DockerTrainStateSchema
    ) -> DockerTrainState:
        db_state = self.read_train_state(db, train_id)
        if not db_state:
            raise HTTPException(
                status_code=404, detail=f"Train State for train: {train_id} not found"
            )
        db_state.num_executions = state_in.num_executions
        db_state.last_execution = state_in.last_execution
        db_state.status = state_in.status

        db.commit()
        db.refresh(db_state)

        return db_state

    def get_train_executions(self, db: Session, train_id: str) -> DockerTrainExecution:
        db_train = self.get_by_train_id(db, train_id)
        if not db_train:
            raise HTTPException(status_code=404, detail=f"Train {train_id} not found")
        executions = db_train.executions
        return executions

    def get_executions(
        self, db: Session, skip: int = 0, limit: int = 100
    ) -> List[DockerTrainExecution]:
        return (
            db.query(DockerTrainExecution)
            .order_by(DockerTrainExecution.start.desc())
            .offset(skip)
            .limit(limit)
            .all()
        )

    def synchronize_central(self, db: Session) -> List[DockerTrain]:

        central_trains = clients.central.get_trains(settings.config.station_id)
        train_objects = []
        for train in central_trains["data"]:
            if train["approval_status"] == "approved":
                db_train = self._parse_central_api_train(db, train_dict=train)
                if db_train:
                    train_objects.append(db_train)
        if train_objects:
            trains, train_states = zip(*train_objects)
        else:
            return []
        return trains

    def _parse_central_api_train(
        self, db: Session, train_dict: dict
    ) -> Union[None, Tuple[DockerTrain, DockerTrainState]]:
        db_train = self.get_by_train_id(db, train_dict["train_id"])
        if db_train:
            # todo update existing train
            return None
        else:
            db_train = self._db_train_from_central_api(train_dict)
            db.add(db_train)
            db.commit()
            db.refresh(db_train)

            db_state = self._train_state_from_central(db_train.id, train_dict)
            db.add(db_state)
            db.commit()
            db.refresh(db_state)

            return db_train, db_state

    def _db_train_from_central_api(self, train_dict: dict) -> DockerTrain:

        image_name = self._make_train_image_name(train_dict)
        proposal_link = self._make_train_proposal_link(train_dict)
        db_train = DockerTrain(
            train_id=train_dict["train_id"],
            created_at=parser.parse(train_dict["created_at"]),
            updated_at=parser.parse(train_dict["updated_at"]),
            proposal=proposal_link,
            type=train_dict["train"]["type"],
            name=train_dict["train"]["name"],
            num_participants=train_dict["train"]["stations"],
            image_name=image_name,
        )
        return db_train

    @staticmethod
    def _make_train_image_name(train_dict: dict) -> str:

        tag = train_dict["artifact_tag"]
        if not tag:
            tag = "latest"
        image_name = "{}/{}/{}:{}".format(
            settings.config.registry.address,
            settings.config.registry.project,
            train_dict["train_id"],
            tag,
        )
        return image_name

    @staticmethod
    def _train_state_from_central(train_id: int, train_dict: dict) -> DockerTrainState:
        state = DockerTrainState(
            train_id=train_id,
            central_status=train_dict["run_status"],
        )
        return state

    @staticmethod
    def _make_train_proposal_link(train_dict: dict):
        proposal_id = train_dict["train"]["proposal_id"]

        central_path = settings.config.central_ui.api_url
        if central_path.endswith("/api"):
            central_path = central_path[:-4]
        proposal_link = f"{central_path}/proposals/{proposal_id}"
        return proposal_link


docker_trains = CRUDDockerTrain(DockerTrain)



import json
from typing import List, Union

import pandas as pd
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from station.app.datasets.filesystem import get_file
from station.app.models.datasets import DataSet
from station.app.schemas.datasets import DataSetCreate, DataSetStatistics, DataSetUpdate

from .base import CreateSchemaType, CRUDBase, ModelType


class CRUDDatasets(CRUDBase[DataSet, DataSetCreate, DataSetUpdate]):

    # TODO fix MinIO Client connection
    # using the .create function from the base CRUD operators
    def create(self, db: Session, *, obj_in: CreateSchemaType) -> ModelType:
        obj_in_data = jsonable_encoder(obj_in)
        db_obj = self.model(**obj_in_data)
        # TODO check for multiple files
        # try:
        #     file = get_file(db_obj.access_path, db_obj.storage_type)

        db.add(db_obj)
        db.commit()
        db.refresh(db_obj)

        return db_obj

    def get_data(
        self, db: Session, data_set_id: str, file_name: Union[str, List[str]] = None
    ):
        dataset = self.get(db, data_set_id)
        if dataset.data_type == "image":
            raise NotImplementedError
        elif dataset.data_type == "csv":
            path = dataset.access_path
            file = get_file(path, dataset.storage_type)
            with file as f:
                csv_df = pd.read_csv(f)
                return csv_df
        elif dataset.data_type == "directory":
            raise NotImplementedError
        elif dataset.data_type == "fhir":
            raise NotImplementedError
        return dataset

    def get_by_name(self, db: Session, name: str):
        dataset = db.query(self.model).filter(self.model.name == name).first()
        return dataset

    def add_stats(
        self,
        db: Session,
        data_set_id: str,
        stats: DataSetStatistics,
        file_name: str = None,
    ):
        dataset = self.get(db, data_set_id)

        if file_name:
            if dataset.summary:
                stored_stats = json.loads(dataset.summary)
            else:
                stored_stats = {}
            stored_stats[file_name] = stats
            dataset.summary = jsonable_encoder(stored_stats)
        else:
            stats = jsonable_encoder(stats)
            stats_json = json.dumps(stats)
            dataset.summary = stats_json
        db.commit()
        db.refresh(dataset)
        return dataset


datasets = CRUDDatasets(DataSet)



# Import all the models, so that Base has them before being
# imported by Alembic
# ruff: noqa
from station.app.db.base_class import Base  # noqa

from station.app.models.docker_trains import (
    DockerTrainConfig,
    DockerTrain,
    DockerTrainState,
    DockerTrainExecution,
)  # noqa
from station.app.models.datasets import DataSet  # noqa
from station.app.models.fhir_server import FHIRServer  # noqa
from station.app.models.discovery import DataSetSummary  # noqa
from station.app.models.notification import Notification
from station.app.models.local_trains import (
    LocalTrain,
    LocalTrainConfig,
    LocalTrainState,
    LocalTrainExecution,
    LocalTrainMasterImage,
)  # noqa

# from station.app.models.user import User  # noqa



import os

import pytest
from dotenv import find_dotenv, load_dotenv

from station.app.schemas.local_trains import LocalTrainCreate
from station.common.clients.station import StationAPIClient


@pytest.fixture
def station_client():
    load_dotenv(find_dotenv())
    base_url = os.getenv("STATION_API_URL", "http://localhost:8000/api")
    auth_url = os.getenv("STATION_AUTH_URL", "http://localhost:3010/token")
    username = os.getenv("STATION_USER")
    password = os.getenv("STATION_PASSWORD")
    return StationAPIClient(
        base_url=base_url, auth_url=auth_url, username=username, password=password
    )


"""
Test local train client

"""


def test_local_train_create(station_client):
    local_train = station_client.local_trains.create(
        LocalTrainCreate(
            name="test-create",
            command="python3 -m station.app.main",
            custom_image="station:latest",
        )
    )
    assert local_train.name == "test-create"
    assert local_train.command == "python3 -m station.app.main"


def test_get_local_train(station_client):
    local_train = station_client.local_trains.get("test-create")
    assert local_train.name == "test-create"
    assert local_train.command == "python3 -m station.app.main"


def test_list_local_trains(station_client):
    local_trains = station_client.local_trains.get_multi()
    print(local_trains)
    assert len(local_trains) > 0


def test_get_local_train_archive(station_client):
    archive = station_client.local_trains.download_train_archive(
        "5b2682dd-b4b1-46af-b2ed-d6aba8a309bb"
    )
    print(archive.getmembers())
    assert archive



from abc import ABC
from typing import Generator, Iterator, List

import numpy as np
import torch
from minio import datatypes
from PIL import Image
from torch.utils.data import DataLoader, Dataset, IterableDataset
from torch.utils.data.dataset import T_co
from torchvision.transforms import CenterCrop, Compose, ToTensor

from station.common.clients.minio import MinioClient


class BaseDataSet(Dataset, ABC):
    id: str


class MinioFolderDataSet(IterableDataset):
    def __init__(
        self, minio_client: MinioClient, id: str = None, transform: Compose = None
    ):
        super(MinioFolderDataSet, self).__init__()
        self.id = id
        self.client = minio_client
        self.transform = transform

    def __iter__(self) -> Iterator[T_co]:
        worker_info = torch.utils.data.get_worker_info()

        if worker_info is None:
            return self._make_generator()

        else:
            print(worker_info.num_workers)
            # TODO
            raise NotImplementedError(
                "Implement sliced generators for multiworker loading"
            )

    def _make_generator(self, id: str = None):
        if id:
            self.id = id
        minio_items = self._get_minio_items()

        tensor_generator = self._make_image_generator(minio_items)
        return tensor_generator

    def _get_minio_items(self) -> Generator[datatypes.Object, None, None]:
        items = self.client.get_minio_dir_items(self.id)
        return items

    def _make_image_generator(
        self, minio_items: Generator[datatypes.Object, None, None]
    ) -> Generator[torch.Tensor, None, None]:
        for i, object in enumerate(minio_items):
            minio_object = self.client.client.get_object(
                "data-sets", object.object_name
            )
            np_img = np.array(Image.open(minio_object))

            print(object.object_name)

            if self.transform:
                yield self.transform(np_img)
            else:
                yield np_img


class MinioFolderDS(Dataset):
    # TODO generalized target creation

    def __init__(
        self,
        client: MinioClient,
        data_set_id: str = None,
        transform: Compose = None,
        target_classes: List[str] = None,
    ):
        super().__init__()
        self.data_set_id = data_set_id
        self.minio_client = client
        self.transform = transform
        self.items = None

        self.get_minio_items()

        if target_classes:
            self.classes = np.asarray(target_classes)
        else:
            self.classes = self.get_classes_from_folders()

    def __getitem__(self, index) -> T_co:
        minio_object = self.minio_client.client.get_object(
            "datasets", self.items[index]
        )
        np_img = np.array(Image.open(minio_object))
        np_img = np.transpose(np_img.astype(float), (2, 1, 0))
        label = self.get_label(self.items[index])
        if self.transform:
            return self.transform(np_img), label
        return torch.Tensor(np_img), label

    def __len__(self):
        return len(list(self.items))

    def get_label(self, item_name: str):
        folder = item_name[len(self.data_set_id) :].split("/")[0]
        label = self.classes == folder
        return torch.tensor(label, dtype=torch.long)

    def get_minio_items(self):
        minio_items = self.minio_client.get_minio_dir_items(self.data_set_id)
        item_list = []
        for item in minio_items:
            item_list.append(item.object_name)

        self.items = item_list

    def save(self):
        # TODO Save the full data set to disk somewhere to enable fast loading
        pass

    def get_classes_from_folders(self):
        classes = self.minio_client.get_classes_by_folders(self.data_set_id)
        print(classes)
        return np.asarray(classes)


if __name__ == "__main__":
    minio_client = MinioClient(
        minio_server="localhost:9000",
        secret_key="minio_admin",
        access_key="minio_admin",
    )

    transform = Compose([ToTensor(), CenterCrop(size=24)])
    minio_ds = MinioFolderDS(
        client=minio_client, data_set_id="cifar/batch_2", transform=transform
    )
    minio_ds.get_minio_items()
    print(len(minio_ds))

    dl = DataLoader(minio_ds, batch_size=6, num_workers=0, shuffle=True)

    for i, batch in enumerate(dl):
        x, y = batch
        print(y)
        if i > 3:
            break



import os
from typing import Any

from dotenv import find_dotenv, load_dotenv
from sqlalchemy.orm import Session

from station.app.crud import federated_trains
from station.common.clients.conductor import ConductorRESTClient
from station.common.clients.minio import MinioClient
from station.worker import SessionLocal
from station.worker.discovery import perform_discovery
from station.worker.loader import BaseLoader, MinioFolderDS, ModelLoader
from station.worker.trainer import FederatedTrainer


class PHTWorker:
    def __init__(self):
        self.conductor_client = ConductorRESTClient()
        self.minio_client = MinioClient()

    def perform_discovery(self, db: Session, train_id: Any):
        db_train = federated_trains.get_by_train_id(db=db, train_id=train_id)
        assert db_train.dataset
        discovery_result = perform_discovery(db_train.dataset)

        discovery_post = {
            "station_id": os.getenv("STATION_ID"),
            "results": discovery_result,
        }
        self.conductor_client.post_discovery_results(
            train_id=train_id, discovery_results=discovery_post
        )

    def make_data_loader(self, db: Session, train_id: Any):
        db_train = federated_trains.get_by_train_id(db=db, train_id=train_id)
        assert db_train.dataset
        # TODO get configuration setting on whether to load and store the data set/loader
        ds = MinioFolderDS(self.minio_client, data_set_id=db_train.dataset.access_path)
        loader = BaseLoader(data_set=ds).make_data_loader()

        return loader

    def train_model(self, db: Session, train_id: Any):
        model = ModelLoader().load_train_model(db, train_id)
        print(model)
        data_loader = self.make_data_loader(db, train_id)
        trainer = FederatedTrainer(gpus=1, max_epochs=1)
        trainer.fit(model=model, train_dataloader=data_loader)

    def distribute_model(self, db: Session, train_id: Any):
        pass


if __name__ == "__main__":
    load_dotenv(find_dotenv())
    session = SessionLocal()
    TRAIN_ID = "1"
    worker = PHTWorker()
    # worker.perform_discovery(session, TRAIN_ID)
    # loader = worker.make_data_loader(session, TRAIN_ID)
    worker.train_model(session, TRAIN_ID)

    # for i, batch in enumerate(loader):
    #     x, y = batch
    #     print(y, x)
    #     print(x.shape)
    #     if i > 3:
    #         break



import datetime
from pathlib import Path

import click
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


def copy_certificates(ctx):
    """
    Copy certificates from the paths specified in the config to the installation directory
    Args:
        ctx: cli context

    Returns:

    """
    click.echo("Copying certificates to installation directory")
    certs = ctx.obj["https"]["certs"][0]
    key_path = certs["key"]
    cert_path = certs["cert"]
    with open(key_path, "rb") as f:
        key = f.read()
    with open(cert_path, "rb") as f:
        cert = f.read()

    dest_path = Path(ctx.obj["install_dir"]) / "certs"
    with open(dest_path / "key.pem", "wb") as f:
        f.write(key)
    with open(dest_path / "cert.pem", "wb") as f:
        f.write(cert)


def generate_certificates(
    domain: str, key_path: str, cert_path: str, key_password: str = None
):
    # Generate our key
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )
    # Write our key to disk for safe keeping
    with open(key_path, "wb") as f:
        f.write(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.BestAvailableEncryption(
                    key_password.encode()
                )
                if key_password
                else serialization.NoEncryption(),
            )
        )

    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, "DE"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Baden-Wuerttemberg"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "Tuebingen"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "PHT"),
            x509.NameAttribute(NameOID.COMMON_NAME, "{}".format(domain)),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(
            # Our certificate will be valid for 1 year
            datetime.datetime.utcnow()
            + datetime.timedelta(days=365)
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("station.local")]),
            critical=False,
            # Sign our certificate with our private key
        )
        .sign(key, hashes.SHA256())
    )
    # Write our certificate out to disk.
    with open(cert_path, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))



import os

from cryptography.fernet import Fernet
from pydantic import SecretStr, validator

from station.common.constants import DefaultValues


def is_default(value: str, default: DefaultValues) -> bool:
    """Validate that the given value is not the default value

    Args:
        value: the value to check
        default: the default value to check against

    Returns:
        True if the value is the default value, False otherwise
    """
    return value == default.value


def validate_fernet_key(value: SecretStr):
    """Validation function for fernet key

    Args:
        value: the fernet key to validate

    Raises:
        ValueError: If the key is empty
        AssertionError: If the key is less than 32 characters or is the default key or is invalid
    Returns:
        the validated key
    """
    if not value:
        raise ValueError("Fernet key must not be empty")
    if len(value.get_secret_value()) < 32:
        raise AssertionError("Fernet key must be at least 32 characters")
    if is_default(value.get_secret_value(), DefaultValues.FERNET_KEY):
        raise AssertionError(
            f"Fernet key must not be the default key [{DefaultValues.FERNET_KEY.value}]"
        )

    # try to load the key to make sure it is valid
    try:
        Fernet(value.get_secret_value())
    except Exception as e:
        raise AssertionError(f"Invalid Fernet key: {e}")
    return value


def validate_admin_password(value: str):
    """Validation function for admin password

    Args:
        value: the admin password to validate

    Raises:
        ValueError: If the password is empty
        AssertionError: If the password is less than 8 characters or is the default password
    Returns:
        the validated password
    """
    if not value:
        raise ValueError("Password must not be empty")
    if len(value) < 8:
        raise AssertionError("Password must be at least 8 characters")
    if is_default(value, DefaultValues.ADMIN_PASSWORD):
        raise AssertionError(
            f"Password must not be the default password [{DefaultValues.ADMIN_PASSWORD.value}]"
        )
    return value


def validate_file_readable(value: str):
    """Validation function for a file that should be readable

    Args:
        value: the path to the file to validate

    Raises:
        AssertionError: if the file is not readable

    Returns:
        the file path
    """
    try:
        with open(value, "r"):
            pass
    except Exception as e:
        raise AssertionError(f"Could not read file {value}: {e}")
    return value


def validate_is_dir_and_writable(value: str):
    if os.path.isdir(value):
        if not os.access(value, os.R_OK):
            raise AssertionError(f"Path {value} is not readable")

        if not os.access(value, os.W_OK):
            raise AssertionError(f"Path {value} is not writable")

        if not os.access(value, os.X_OK):
            raise AssertionError(f"Path {value} is not executable")

        return value

    raise AssertionError(f"Path {value} is not a directory")


def file_readable_validator(field_name: str):
    """Reusable pydantic validator function for a file that should be readable"""
    return validator(field_name, allow_reuse=True)(validate_file_readable)


def dir_validator(field_name: str):
    """
    Checks wether a path is a directory and is writable
    """
    return validator(field_name, allow_reuse=True)(validate_is_dir_and_writable)


def admin_validator(field_name: str = "admin_password"):
    """Reusable pydantic validator function for admin password"""
    return validator(field_name, allow_reuse=True)(validate_admin_password)



import os

import pytest
from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


@pytest.fixture()
def train_id():
    return "airflow_test"


def test_run(train_id):
    if os.getenv("ENVIRONMENT") == "testing":
        response = client.post(
            "/api/trains/docker",
            json={
                "train_id": train_id,
                "config": {
                    "name": "testconfig",
                    "airflow_config": {
                        "repository": f"dev-harbor.grafm.de/station_1/{train_id}",
                        "tag": "latest",
                    },
                },
            },
        )
        if response.status_code == 200:

            run_pht_train = client.post(
                "/api/airflow/run_pht_train/run", json={"train_id": train_id}
            )
            response_pht_train = run_pht_train.json()

            assert run_pht_train.status_code == 200
            assert response_pht_train["dag_id"] == "run_pht_train"
            assert response_pht_train["train_id"] == "airflow_test"
            assert response_pht_train["start_date"]

            execution = client.get(f"/api/trains/docker/{train_id}/executions")
            assert (
                execution.json()[-1]["airflow_dag_run"] == response_pht_train["run_id"]
            )


def test_run_fails(train_id):
    if os.getenv("ENVIRONMENT") == "testing":
        run_wrong_dag = client.post(
            "/api/airflow/dag_not_there/run", json={"train_id": train_id}
        )
        assert run_wrong_dag.status_code == 404, run_wrong_dag.text


def test_get_airflow_run_information(train_id):
    if os.getenv("ENVIRONMENT") == "testing":
        response = client.get(f"/api/trains/docker/{train_id}/executions")
        executions = response.json()
        assert len(executions) > 0
        run_id = executions[-1]["airflow_dag_run"]

        run_info_response = client.get(f"/api/airflow/logs/run_pht_train/{run_id}")
        print(run_info_response.json())
        assert run_info_response.status_code == 200


def test_get_airflow_run_information_fails():
    if os.getenv("ENVIRONMENT") == "testing":
        with pytest.raises(Exception):
            run_id_fail = client.get("/api/airflow/logs/run_pht_train/no_run_id")
            assert run_id_fail.status_code == 404

        with pytest.raises(Exception):
            dag_id_fail = client.get("/api/airflow/logs/dag_not_there/no_run_id")
            assert dag_id_fail.status_code == 404


def test_get_airflow_task_log(train_id):
    if os.getenv("ENVIRONMENT") == "testing":
        response = client.get(f"/api/trains/docker/{train_id}/executions")
        executions = response.json()
        assert len(executions) > 0
        run_id = executions[-1]["airflow_dag_run"]

        run_info_response = client.get(f"/api/airflow/logs/run_pht_train/{run_id}")
        tasklist = run_info_response.json()["tasklist"]
        for task in tasklist["task_instances"]:
            task_id = task["task_id"]
            task_info = client.get(
                f"/api/airflow/logs/run_pht_train/{run_id}/{task_id}"
            )
            assert task_info.json()["run_info"]
            assert task_info.status_code == 200

        # with given task try number
        for task in tasklist["task_instances"]:
            task_id = task["task_id"]
            task_info = client.get(
                f"/api/airflow/logs/run_pht_train/{run_id}/{task_id}/?task_try_number=0"
            )
            assert task_info.json()["run_info"]
            assert task_info.status_code == 200


def test_get_airflow_task_log_fails(train_id):
    if os.getenv("ENVIRONMENT") == "testing":
        response = client.get(f"/api/trains/docker/{train_id}/executions")
        executions = response.json()
        assert len(executions) > 0
        run_id = executions[-1]["airflow_dag_run"]

        task_info = client.get(f"/api/airflow/logs/run_pht_train/{run_id}/no_task_id")
        assert task_info.status_code == 404



import pytest
from dotenv import find_dotenv, load_dotenv
from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.crud.crud_fhir_servers import fhir_servers
from station.app.main import app
from station.app.schemas.fhir import FHIRServerCreate
from station.app.settings import settings

from .test_db import TestingSessionLocal, override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


@pytest.fixture()
def station_settings():
    load_dotenv(find_dotenv())
    settings.setup()
    return settings


def test_fhir_server_encrypted_storage(station_settings):
    load_dotenv(find_dotenv())

    db = TestingSessionLocal()

    obj_in = {
        "name": "Test Server",
        "api_address": "http://test.com",
        "username": "user",
        "password": "password",
    }
    obj_in = FHIRServerCreate(**obj_in)
    fhir_server = fhir_servers.create(db, obj_in=obj_in)

    station_fernet = station_settings.get_fernet()
    assert fhir_server.name == "Test Server"
    assert fhir_server.password != "password"
    assert station_fernet.decrypt(fhir_server.password.encode()) == b"password"

    db.close()


def test_fhir_server_create():
    """
    Test the creation of a FHIR server
    """
    response = client.post(
        "/api/fhir/server",
        json={"name": "Test Server", "api_address": "http://test.com"},
    )
    assert response.status_code == 201
    assert response.json()["name"] == "Test Server"
    assert response.json()["api_address"] == "http://test.com"


def test_fhir_server_get():
    """
    Test the retrieval of a FHIR server
    """
    response = client.get("/api/fhir/server/1")
    assert response.status_code == 200
    assert response.json()["name"] == "Test Server"
    assert response.json()["api_address"] == "http://test.com"


def test_list_fhir_servers():
    """
    Test the retrieval of a list of FHIR servers
    """
    response = client.post(
        "/api/fhir/server",
        json={"name": "Test Server", "api_address": "http://test.com"},
    )
    assert response.status_code == 201
    assert response.json()["name"] == "Test Server"
    assert response.json()["api_address"] == "http://test.com"
    response = client.get("/api/fhir/server")
    assert response.status_code == 200
    assert response.json()[0]["name"] == "Test Server"
    assert response.json()[0]["api_address"] == "http://test.com"


def test_update_fhir_server():
    """
    Test the update of a FHIR server
    """
    response = client.put(
        "/api/fhir/server/1",
        json={
            "name": "Test Server Updated",
            "api_address": "http://test.com",
            "username": "user",
            "password": "password",
        },
    )
    assert response.status_code == 200
    assert response.json()["name"] == "Test Server Updated"
    assert response.json()["api_address"] == "http://test.com"


def test_delete_fhir_server():
    """
    Test the deletion of a FHIR server
    """
    response = client.delete("/api/fhir/server/1")
    assert response.status_code == 202



from typing import Optional

from fhir_kindling.fhir_query.query_parameters import FHIRQueryParameters
from pydantic import BaseModel
from sqlalchemy.orm import Session

from station.app.crud.crud_datasets import datasets
from station.app.crud.crud_local_train import local_train
from station.app.crud.crud_train_configs import docker_train_config
from station.app.crud.local_train_master_image import local_train_master_image
from station.app.schemas.local_trains import LocalTrainExecution
from station.common.clients.airflow.client import airflow_client
from station.common.clients.airflow.docker_trains import (
    process_dataset,
    process_db_config,
)


class FHIRConfig(BaseModel):
    server_id: str
    query_string: Optional[str] = None
    parameters: Optional[FHIRQueryParameters] = None


class AirflowRunConfig(BaseModel):
    train_id: str
    env: Optional[dict] = None
    volumes: Optional[dict] = None
    master_image: Optional[str] = None
    custom_image: Optional[str] = None
    fhir: Optional[FHIRConfig] = None


def run_local_train(
    db: Session, train_id: str, dataset_id: str = None, config_id: int = None
) -> LocalTrainExecution:
    """
    Run a local train

    :param train_id: id of the train
    :param dataset_id: optional id of the dataset
    :param config_id: optional id of the config
    :return: config dictionary
    """
    db_train = local_train.get(db, train_id)
    if db_train is None:
        raise ValueError(f"Train {train_id} not found")
    config = make_dag_config(db, db_train, train_id, dataset_id, config_id)
    print(config)
    run_id = airflow_client.trigger_dag("run_local_train", config=config)
    train_execution = local_train.create_run(
        db,
        train_id=train_id,
        dag_run=run_id,
        config_id=config_id,
        dataset_id=dataset_id,
    )

    return train_execution


def make_dag_config(
    db: Session, db_train, train_id: str, dataset_id: str = None, config_id: int = None
) -> dict:
    """
    Create a config dictionary for the airflow DAG

    :param train_id: id of the train
    :param dataset_id: optional id of the dataset
    :param config_id: optional id of the config
    :return: config dictionary
    """

    if db_train.master_image_id:
        master_image = local_train_master_image.get(
            db, db_train.master_image_id
        ).image_id
    else:
        master_image = None

    dag_config = {
        "train_id": train_id,
        "master_image": master_image,
        "custom_image": db_train.custom_image,
    }

    if dataset_id:
        ds = datasets.get(db, dataset_id)
        process_dataset(dag_config, ds)
    if config_id:
        db_config = docker_train_config.get(db, config_id)
        process_db_config(dag_config, db_config)

    return dag_config



import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from conductor_lib.src.torch import LightningTrainModel
from torchvision import transforms


class Cifar10Model(LightningTrainModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)

        loss = F.cross_entropy(y_hat, y)
        return loss

    def configure_optimizers(self):
        return optim.SGD(self.parameters(), lr=0.001, momentum=0.9)

    def make_transform(self) -> transforms.Compose:
        transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )

        return transform



import psutil
from fastapi import APIRouter

from station.app.config import clients
from station.app.schemas import station_status as status_schema

router = APIRouter()
"""
The station status  endpoint returns the status of local and global components  (fhir  airflow harbo minio
"""


def service_health_check():
    """
    Get the health status of all connected services
    """
    service_status = []
    services = {
        "airflow": clients.airflow.health_check(),
        "harbor": clients.harbor.health_check(),
        "minio": clients.minio.health_check(),
    }
    for service, health in services.items():
        service_status.append(status_schema.ServiceStatus(name=service, status=health))
    return service_status


def get_hardware_resources_status():
    # get memory statistics
    memory_stats = psutil.virtual_memory()
    memory_usage = status_schema.MemoryUsage(
        total=memory_stats.total,
        available=memory_stats.available,
        percent=memory_stats.percent,
        used=memory_stats.used,
        free=memory_stats.free,
    )

    # get disk statistics
    disk_util = psutil.disk_usage("./")
    disk_usage = status_schema.DiskUsage(
        total=disk_util.total,
        used=disk_util.used,
        free=disk_util.free,
        percent=disk_util.percent,
    )

    # cpu usage per core
    cpu_usage = psutil.cpu_percent(interval=None, percpu=True)

    # todo GPU
    return status_schema.HardwareResources(
        memory=memory_usage,
        cpu=cpu_usage,
        disk=disk_usage,
    )


@router.get("", response_model=status_schema.StationStatus)
async def get_station_status():
    hardware = get_hardware_resources_status()
    services = service_health_check()

    return status_schema.StationStatus(hardware=hardware, services=services)


# @router.get("/container_resource_util")
# def status_docker_container_resource_use():
#     """
#     get information for all docker containers
#     """
#     return dockerClient.get_stats_all()
#
#
# @router.get("/container/{container_id}")
# def status_docker_container_resource_use(container_id: Any):
#     """
#     get information for container id
#     """
#     return dockerClient.get_stats_container(container_id)
#
#
# @router.get("/container_info")
# def container_info():
#     return dockerClient.get_information_all()



from enum import Enum


class StationEnvironmentVariables(Enum):
    # station configuration variables
    STATION_ID = "STATION_ID"
    ENVIRONMENT = "ENVIRONMENT"
    STATION_DB = "STATION_DB"
    STATION_API_HOST = "STATION_API_HOST"
    STATION_API_PORT = "STATION_API_PORT"
    FERNET_KEY = "FERNET_KEY"
    STATION_DATA_DIR = "STATION_DATA_DIR"
    CONFIG_PATH = "STATION_CONFIG_PATH"

    # central api configuration variables
    CENTRAL_API_URL = "CENTRAL_API_URL"
    STATION_ROBOT_ID = "STATION_ROBOT_ID"
    STATION_ROBOT_SECRET = "STATION_ROBOT_SECRET"

    # airflow environment variables
    AIRFLOW_HOST = "AIRFLOW_HOST"
    AIRFLOW_PORT = "AIRFLOW_PORT"
    AIRFLOW_USER = "AIRFLOW_USER"
    AIRFLOW_PW = "AIRFLOW_PW"
    AIRFLOW_API_URL = "AIRFLOW_API_URL"

    # Redis environment variables
    REDIS_HOST = "REDIS_HOST"
    REDIS_PORT = "REDIS_PORT"
    REDIS_PW = "REDIS_PW"
    REDIS_DB = "REDIS_DB"

    # auth environment variables
    AUTH_SERVER_HOST = "AUTH_SERVER_HOST"
    AUTH_SERVER_PORT = "AUTH_SERVER_PORT"
    AUTH_ROBOT_ID = "AUTH_ROBOT_ID"
    AUTH_ROBOT_SECRET = "AUTH_ROBOT_SECRET"

    # Registry environment variables
    REGISTRY_URL = "HARBOR_URL"
    REGISTRY_USER = "HARBOR_USER"
    REGISTRY_PW = "HARBOR_PW"
    REGISTRY_PROJECT = "HARBOR_PROJECT"

    # minio environment variables
    MINIO_HOST = "MINIO_HOST"
    MINIO_PORT = "MINIO_PORT"
    MINIO_ACCESS_KEY = "MINIO_USER"
    MINIO_SECRET_KEY = "MINIO_PW"



import urllib.parse

import pendulum
import requests
from authup import Authup
from pydantic import SecretStr


class BaseClient:
    def __init__(
        self,
        base_url: str,
        auth_url: str = None,
        username: str = None,
        password: str = None,
        robot_id: str = None,
        robot_secret: str = None,
        headers: dict = None,
    ):
        self.base_url = base_url
        self.auth_url = auth_url
        self.username = username
        self.password = password
        self.robot_id = robot_id
        self.robot_secret = robot_secret
        self.token = None
        self.refresh_token = None
        self.token_expiration = None
        self._headers = headers

        self.auth = Authup(
            url=self.auth_url,
            username=self.username,
            password=self.password,
            robot_id=self.robot_id,
            robot_secret=self.robot_secret,
        )

        if not self.auth_url:
            self.auth_url = f"{self.base_url}/auth"

        # self.setup()

    @property
    def headers(self) -> dict:
        return self.auth.get_authorization_header()

    # def setup(self):
    #     self._get_token()

    def _get_token(self) -> str:
        if not self.token or self.token_expiration < pendulum.now():
            if self.username and self.password:
                r = requests.post(
                    self.auth_url,
                    data={"username": self.username, "password": self.password},
                )
            elif self.robot_id and self.robot_secret:
                if isinstance(self.robot_secret, SecretStr):
                    self.robot_secret = self.robot_secret.get_secret_value()
                r = requests.post(
                    self.auth_url,
                    data={"id": self.robot_id, "secret": self.robot_secret},
                )
            else:
                raise Exception("No credentials provided")

            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                print(r.text)
                raise e
            r = r.json()
            self.token = r["access_token"]
            self.token_expiration = pendulum.now().add(seconds=r["expires_in"])

        return self.token

    @staticmethod
    def _make_url_safe(url: str) -> str:
        return urllib.parse.quote(url, safe="=&?")



import copy
import re

import pytest
from rich.console import Console

from station.common.constants import CERTS_REGEX
from station.ctl.config import validate_config


@pytest.fixture
def config_dict():
    config = {
        "id": "d7b7bd69-c828-45f3-b0bc-0d5ca10a8cd5",
        "version": "latest",
        "environment": "development",
        "admin_password": "GucVuG6MgVyy58v8Xjg3o4jTnAyNrP1k",
        "data_dir": "./station_data",
        "central": {
            "api_url": "https://dev.personalhealthtrain.de/api",
            "robot_id": "0b34acb9-9b26-4780-80d1-705772464cf2",
            "robot_secret": "start123",
            "private_key": "./private.pem",
            "private_key_password": "",
        },
        "http": {"port": "80"},
        "https": {
            "port": "443",
            "domain": "station.localhost",
            "certificate": {
                "cert": "./certs/cert.pem",
                "key": "./certs/cert.pem",
            },
        },
        "traefik": {"dashboard": True, "dashboard_port": 8081},
        "registry": {
            "address": "dev-harbor.personalhealthtrain.de",
            "user": "test-user",
            "password": "test-password",
            "project": "test-project",
        },
        "db": {"host": "127.0.0.1", "admin_user": "admin", "admin_password": "admin"},
        "api": {"fernet_key": "Z_kebTcA7p2VV9xga-ES2wCMjvfaRNzQktjsxo5vPMM="},
        "airflow": {
            "host": "127.0.0.1",
            "admin_user": "admin",
            "admin_password": "start123",
        },
        "auth": {
            "admin_user": "admin",
            "host": "127.0.0.1",
            "port": 3001,
            "robot_id": "5f77fe3f-a48c-46be-ba67-df6b1058ebcb",
            "robot_secret": "v1ziczynshyotuzudc8ymumjtijli9yoo1mygyrv2ucqdm77lae6d5pni6xh5vp4",
        },
        "minio": {
            "host": "127.0.0.1",
            "port": 9000,
            "admin_user": "minio_admin",
            "admin_password": "minio_admin",
        },
        "redis": {
            "host": "127.0.0.1",
        },
    }
    return config


def test_validate_config_object(config_dict):
    test_dict = copy.deepcopy(config_dict)

    # test_dict["id"] = None
    test_dict["airflow"]["host"] = None
    test_dict["https"]["certificate"]["key"] = None
    test_dict["admin_password"] = "password"
    results = validate_config(test_dict)

    for result in results:
        print(result)


def test_validate_config(config_dict):
    results, table = validate_config(config_dict)
    console = Console()
    print()
    console.print(table)


def test_match_cert_index():
    field_0 = "https.certs[0]"
    field_1 = "https.certs[12]"
    field_2 = "htps.certs[0]"
    field_3 = "https.certs[a]"

    assert re.match(CERTS_REGEX, field_0)
    assert re.match(CERTS_REGEX, field_0).group(1) == "0"
    assert re.match(CERTS_REGEX, field_1)
    assert re.match(CERTS_REGEX, field_2) is None
    assert re.match(CERTS_REGEX, field_3) is None



from datetime import datetime

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.auth import authorized_user

# from station.common.clients.airflow.client import airflow_client
from station.app.config import clients
from station.app.crud.crud_docker_trains import docker_trains
from station.app.schemas.airflow import (
    AirflowInformation,
    AirflowRun,
    AirflowRunMsg,
    AirflowTaskLog,
)
from station.app.schemas.docker_trains import DockerTrainExecution
from station.app.schemas.users import User
from station.common.clients.airflow import docker_trains as airflow_docker_train

router = APIRouter()


@router.post("/{dag_id}/run", response_model=AirflowRun)
def run(
    run_msg: AirflowRunMsg,
    dag_id: str,
    db: Session = Depends(dependencies.get_db),
    user: User = Depends(authorized_user),
):
    """
    Trigger a dag run and return the run_id of the run
    @param dag_id: ID of the DAG e.G. "run_local" , "run_pht_train" etc.
    @param run_msg: UID of the train
    @param db:  reference to the postgres database

    Args:
        user:
    """

    if dag_id == "run_local":
        raise NotImplementedError("run_local is not implemented yet")

    elif dag_id == "run_pht_train":
        train = docker_trains.get_by_train_id(db, run_msg.train_id)
        if not train.config_id:
            run_config = None
            config_id = "default"
        else:
            run_config = DockerTrainExecution(config_id=train.config_id)
            config_id = train.config_id
        execution = airflow_docker_train.run_train(db, run_msg.train_id, run_config)
        run_id = execution.airflow_dag_run

    else:
        raise HTTPException(
            status_code=404, detail=f"DAG with id '{dag_id}' not found."
        )

    return {
        "run_id": run_id,
        "config_id": config_id,
        "dag_id": dag_id,
        "train_id": run_msg.train_id,
        "start_date": datetime.now(),
    }


@router.get("/logs/{dag_id}/{run_id}", response_model=AirflowInformation)
def get_airflow_run_information(
    dag_id: str, run_id: str, user: User = Depends(authorized_user)
):
    """
    Get information about one airflow DAG execution.
    @param dag_id: ID of the DAG e.G. "run_local" , "run_pht_train" etc.
    @param run_id: Airflow run ID
    @return:
    """

    run_info = clients.airflow.get_run_information(dag_id, run_id)

    for instance in run_info["tasklist"]["task_instances"][:]:
        try:
            instance.pop("sla_miss")
        except KeyError:
            pass
        try:
            instance.pop("pool_slots")
        except KeyError:
            pass
        try:
            instance.pop("pool")
        except KeyError:
            pass

    return run_info


@router.get(
    "/logs/{dag_id}/{run_id}/{task_id}/{task_try_number}", response_model=AirflowTaskLog
)
def get_airflow_task_log(
    dag_id: str,
    run_id: str,
    task_id: str,
    task_try_number: int,
    user: User = Depends(authorized_user),
):
    """
    Get log of a task in a DAG execution.
    @param dag_id: ID of the DAG e.G. "run_local" , "run_pht_train" etc.
    @param task_id: id of teh task
    @param run_id: Airflow run ID
    @param task_try_number: specific try number for log request
    @return:
    """
    run_info_data = clients.airflow.get_task_log(
        dag_id, run_id, task_id, task_try_number
    )
    if not run_info_data:
        raise HTTPException(status_code=404, detail=f"{task_id} not found.")
    return {"run_info": run_info_data}



import pathlib

import uvicorn
from dotenv import find_dotenv, load_dotenv


def main():
    load_dotenv(find_dotenv())
    # set the path to the project root
    project_root = pathlib.Path(__file__).parent.parent
    uvicorn.run(
        "station.app.main:app",
        host="0.0.0.0",
        port=8000,
        reload=True,
        reload_dirs=[str(project_root)],
        access_log=True,
    )


if __name__ == "__main__":
    main()



from typing import Dict, List, Union

from station.app.models.datasets import DataSet
from station.common.clients.minio import MinioClient


def perform_discovery(ds: DataSet):
    """

    :param ds:
    :return:
    """
    if ds.storage_type == "minio":
        class_dist = _minio_discovery(ds)
        return class_dist
    elif ds.storage_type == "file":
        pass
    else:
        raise NotImplementedError(f"Unsupported storage type: {ds.storage_type}")


def _minio_discovery(ds: DataSet) -> List[Dict[str, Union[int, str]]]:
    print("Performing minio discovery")
    minio_client = MinioClient()

    local_classes = minio_client.get_classes_by_folders(ds.access_path)
    class_distribution = minio_client.get_class_distributions(
        ds.access_path, local_classes
    )

    return class_distribution



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated


class StorageType(Enum):
    """
    Enum for storage types
    """

    LOCAL = "local"
    MINIO = "minio"
    DB = "db"


class DataType(Enum):
    """
    Enum for data types
    """

    IMAGE = "image"
    GENOME = "genome"
    FHIR = "fhir"
    CSV = "csv"
    STRUCTURED = "structured"
    UNSTRUCTURED = "unstructured"
    HYBRID = "hybrid"


class DataSetBase(BaseModel):
    name: str
    data_type: Optional[DataType] = None
    storage_type: Optional[StorageType] = None
    proposal_id: Optional[str] = None
    access_path: Optional[str] = None

    class Config:
        use_enum_values = True


class DataSetCreate(DataSetBase):
    pass


class DataSetUpdate(DataSetBase):
    pass


class MinioFile(BaseModel):
    file_name: str
    full_path: Optional[str] = None
    size: Optional[int] = None
    updated_at: Optional[datetime] = None


class FigureData(BaseModel):
    layout: dict
    data: list


class DataSetFigure(BaseModel):
    fig_data: Optional[FigureData]


class DataSetColumn(BaseModel):
    title: Optional[str]
    not_na_elements: Optional[int]
    figure: Optional[DataSetFigure]


class DataSetUniqueColumn(DataSetColumn):
    type: Literal["unique"]
    number_of_duplicates: Optional[int]


class DataSetEqualColumn(DataSetColumn):
    type: Literal["equal"]
    value: Optional[str]


class DataSetCategoricalColumn(DataSetColumn):
    type: Literal["categorical"]
    number_categories: Optional[int]
    value_counts: Optional[Dict[str, int]]
    most_frequent_element: Optional[Union[int, str]]
    frequency: Optional[int]


class DataSetNumericalColumn(DataSetColumn):
    type: Literal["numeric"]
    mean: Optional[float]
    std: Optional[float]
    min: Optional[float]
    max: Optional[float]


class DataSetStatistics(BaseModel):
    created_at: Optional[datetime] = datetime.now()
    n_items: Optional[int] = 0
    n_features: Optional[int] = 0
    column_information: Optional[
        List[
            Annotated[
                Union[
                    DataSetCategoricalColumn,
                    DataSetNumericalColumn,
                    DataSetEqualColumn,
                    DataSetUniqueColumn,
                ],
                Field(discriminator="type"),
            ]
        ]
    ]

    class Config:
        orm_mode = True


class DataSet(DataSetBase):
    id: Any
    created_at: datetime
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True



from datetime import datetime
from typing import Any, Dict, Union

from sqlalchemy.orm import Session

from station.app.crud import docker_trains
from station.app.models.docker_trains import DockerTrain, DockerTrainConfig
from station.app.schemas.docker_trains import (
    DockerTrainConfigCreate,
    DockerTrainConfigUpdate,
)

from .base import CRUDBase, ModelType, UpdateSchemaType


class CRUDDockerTrainConfig(
    CRUDBase[DockerTrainConfig, DockerTrainConfigCreate, DockerTrainConfigUpdate]
):
    def create(
        self, db: Session, *, obj_in: DockerTrainConfigCreate
    ) -> DockerTrainConfig:
        db_config = DockerTrainConfig(
            name=obj_in.name,
            airflow_config=obj_in.airflow_config.dict()
            if obj_in.airflow_config
            else None,
            gpu_requirements=obj_in.gpu_requirements,
            cpu_requirements=obj_in.cpu_requirements,
            auto_execute=obj_in.auto_execute,
        )
        db.add(db_config)
        db.commit()
        db.refresh(db_config)
        return db_config

    def assign_to_train(
        self, db: Session, train_id: str, config_id: int
    ) -> DockerTrain:
        train = docker_trains.get_by_train_id(db, train_id)
        train.config_id = config_id
        train.updated_at = datetime.now()
        db.commit()
        db.refresh(train)
        return train

    def get_by_name(self, db: Session, name: str) -> DockerTrainConfig:
        config = (
            db.query(DockerTrainConfig).filter(DockerTrainConfig.name == name).first()
        )
        return config

    def update(
        self,
        db: Session,
        *,
        db_obj: ModelType,
        obj_in: Union[UpdateSchemaType, Dict[str, Any]]
    ) -> ModelType:
        obj = super().update(db, db_obj=db_obj, obj_in=obj_in)
        obj.updated_at = datetime.now()
        db.add(obj)
        db.commit()
        db.refresh(obj)
        return obj


docker_train_config = CRUDDockerTrainConfig(DockerTrainConfig)



import os
from typing import List, Union

import requests

from station.app.schemas.local_trains import LocalTrainMasterImageBase
from station.app.schemas.station_status import HealthStatus


class HarborClient:
    def __init__(self, api_url: str = None, username: str = None, password: str = None):
        # Setup and verify connection parameters either based on arguments or .env vars

        self.domain = api_url if api_url else os.getenv("HARBOR_URL")
        self.api_url = self.domain
        if not self.api_url.startswith("https://"):
            self.api_url = "https://" + self.api_url
        if not self.api_url.endswith("/api/v2.0"):
            self.api_url = self.api_url + "/api/v2.0"

        # self.url = self.url.rstrip("/") + "/api/v2.0"

        self.username = username if username else os.getenv("HARBOR_USER")
        assert self.username

        self.password = password if password else os.getenv("HARBOR_PW")
        assert self.password

    def get_artifacts_for_station(
        self, station_id: Union[str, int] = None
    ) -> List[dict]:
        # TODO chache no replys
        if not station_id:
            station_id = int(os.getenv("STATION_ID"))
        assert station_id

        endpoint = f"/projects/station_{station_id}/repositories/"
        r = requests.get(self.api_url + endpoint, auth=(self.username, self.password))
        results = r.json()
        print(results)

        link = True
        while link:
            if r.links:
                if next(iter(r.links)) == "next":
                    print("Getting repositories on next page.")
                    url = r.links["next"]["url"]
                    new_endpoint = url[9:]
                    r = requests.get(
                        self.api_url + new_endpoint, auth=(self.username, self.password)
                    )
                    new_results = r.json()
                    results.extend(new_results)
                else:
                    print("No further repositories found.")
                    link = False
            else:
                # just one page or station not defined
                link = False

        return results

    def get_master_images(self) -> List[LocalTrainMasterImageBase]:
        """
        returns names of master images form harbor
        """
        endpoint = "/projects/master/repositories"
        r = requests.get(self.api_url + endpoint, auth=(self.username, self.password))
        master_images = []
        for repo in r.json():
            project, group, artifact = repo["name"].split("/")
            image_id = f"{self.domain}/{project}/{group}/{artifact}"
            master_images.append(
                LocalTrainMasterImageBase(
                    registry=self.domain,
                    group=group,
                    artifact=artifact,
                    image_id=image_id,
                )
            )

        return master_images

    def health_check(self) -> HealthStatus:
        """
        requests the central service
        @return: dict: status of central harbor instance
        """
        url = self.api_url + "/health"
        try:
            r = requests.get(url=url, auth=(self.username, self.password))
            if r and r.status_code == 200:
                return HealthStatus.healthy
            else:
                return HealthStatus.error
        except requests.exceptions.ConnectionError as e:
            print(e)
        return HealthStatus.error


harbor_client = None



import os
from typing import Generator

from authup.plugins.fastapi import AuthupUser

from station.app.db.session import SessionLocal

# reusable_oauth2 = OAuth2PasswordBearer(
#     tokenUrl=f"{settings.API_V1_STR}/login/access-token"
# )


def get_db() -> Generator:
    try:
        db = SessionLocal()
        yield db
    finally:
        db.close()


def fernet_key() -> bytes:
    # load fernet key from environment variables
    fernet_key = os.getenv("FERNET_KEY")
    if not fernet_key:
        # TODO load key from station config file
        pass
    return fernet_key.encode()


auth_url = os.getenv("AUTH_SERVER_HOST")


authorized_user = AuthupUser(url=auth_url)



import os
from datetime import datetime
from typing import Any, Tuple, Union

from fastapi import HTTPException
from loguru import logger
from sqlalchemy.orm import Session

from station.app.config import clients
from station.app.crud.crud_datasets import datasets
from station.app.crud.crud_docker_trains import docker_trains
from station.app.crud.crud_train_configs import docker_train_config
from station.app.models import docker_trains as dtm
from station.app.schemas import docker_trains as dts
from station.app.settings import settings


def run_train(
    db: Session, train_id: Any, execution_params: dts.DockerTrainExecution
) -> dts.DockerTrainSavedExecution:
    """
    Execute a PHT 1.0 docker train using a configured airflow instance

    :param db: database session
    :param train_id: identifier of the train
    :param execution_params: given config_id or config_json can be used for running train
    :return:
    """
    # Extract the train from the database
    db_train = docker_trains.get_by_train_id(db, train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Train with id '{train_id}' not found."
        )

    # Use default config if there is no config defined.
    if execution_params is None:
        config_id = db_train.config_id
        if not config_id:
            config_id = "default"
        execution_params = dts.DockerTrainExecution(config_id=config_id)

    config_id, config_dict = validate_run_config(db, train_id, execution_params)
    # process assigned data sets
    if execution_params.dataset_id:
        dataset = datasets.get(db, execution_params.dataset_id)
        process_dataset(config_dict, dataset)

    # Execute the train using the airflow rest api
    try:
        run_id = clients.airflow.trigger_dag("run_pht_train", config=config_dict)
        db_train = update_train_after_run(
            db, db_train, run_id, config_id, dataset_id=execution_params.dataset_id
        )
        last_execution = db_train.executions[-1]
        return last_execution
    except Exception as e:
        logger.error(
            f"Error while running train {train_id} with config {config_dict} \n {e}"
        )
        raise HTTPException(
            status_code=503,
            detail="No connection to the airflow client could be established.",
        )


def validate_run_config(
    db: Session,
    train_id: str,
    execution_params: dts.DockerTrainExecution,
    tag: str = None,
) -> Tuple[Union[int, str], dict]:
    """
    Validate the config used for the triggered run
    :param db: database session
    :param train_id: train id of the train to run
    :param execution_params: includes the config_id of the config to use or the specified config
    :param tag: optional tag of the image
    :return:

    """

    harbor_url = settings.config.registry.address
    project = settings.config.registry.project
    config = {
        "repository": f"{harbor_url}/{project}/{train_id}",
        "tag": "latest" if not tag else tag,
    }

    # Extract config by id if given
    if execution_params.config_id != "default":
        db_config = docker_train_config.get(db, execution_params.config_id)
        process_db_config(config, db_config)
        return db_config.id, config
    # Using the default config
    else:
        logger.info(f"Starting train {train_id} using default config")
        # Default config specifies only the identifier of the the train image and uses the latest tag
        return "default", config


def update_state(
    db: Session, db_train: dtm.DockerTrain, run_time: datetime
) -> dts.DockerTrainState:
    """
    Update the train state object of the train after starting an execution.
    Args:
        db: database session
        db_train: train object
        run_time: time when the run was triggered

    Returns:
        updated train state object

    """
    train_state = (
        db.query(dtm.DockerTrainState)
        .filter(dtm.DockerTrainState.train_id == db_train.id)
        .first()
    )
    if train_state:
        train_state.last_execution = run_time
        train_state.num_executions += 1
        train_state.status = "active"
    else:
        logger.info("No train state assigned.")
    db.add(train_state)
    db.commit()
    db.refresh(train_state)

    return train_state


def update_train_after_run(
    db: Session,
    db_train: dtm.DockerTrain,
    run_id: str,
    config_id: int,
    dataset_id: int = None,
) -> dts.DockerTrain:
    """
    Update train in the database and create a new execution object that stores the run configuration.

    Args:
        db: database session
        db_train: train object
        run_id: run id of the airflow DAG run
        config_id: id of the config used for the run
        dataset_id: id of the dataset used for the run

    Returns:
        updated train object
    """
    db_train.is_active = True
    run_time = datetime.now()
    db_train.updated_at = run_time

    # Update the train state
    update_state(db, db_train, run_time)

    # Create an execution
    execution = dtm.DockerTrainExecution(
        train_id=db_train.id,
        airflow_dag_run=run_id,
        config=config_id,
        dataset=dataset_id,
    )
    db.add(execution)
    db.commit()
    db.refresh(execution)

    db.commit()

    return db_train


def process_db_config(config_dict: dict, db_config: dtm.DockerTrainConfig) -> dict:
    """
    Update the config dictionary and with the values from db configuration
    :param config_dict: config dictionary
    :param db_config: db_config object
    :return:
    """

    if db_config.airflow_config:
        db_config = dts.DockerTrainConfig.from_orm(db_config)
        env_dict = {}
        for env in db_config.airflow_config.env:
            env_dict[env.key] = env.value
        config_dict["env"] = env_dict

        volume_dict = {}
        for volume in db_config.airflow_config.volumes:
            volume_dict[volume.host_path] = {
                "bind": volume.container_path,
                "mode": volume.mode,
            }
        config_dict["volumes"] = volume_dict
        return config_dict


def process_dataset(config_dict, dataset):
    """
    Update the config dictionary with the values from the dataset
    Args:
        config_dict:
        dataset:

    Returns:

    """
    mount_path = os.path.join(
        settings.config.station_data_dir, "datasets", str(dataset.id)
    )
    volumes = config_dict.get("volumes", {})
    volumes[mount_path] = {"bind": "/opt/train_data/", "mode": "ro"}
    config_dict["volumes"] = volumes



from datetime import datetime
from typing import Any, Optional

from fhir_kindling.fhir_server.server_responses import ServerSummary
from pydantic import BaseModel, root_validator


class FHIRServerBase(BaseModel):
    """
    Base class for FHIR Server
    """

    api_address: str
    name: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    type: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    oidc_provider_url: Optional[str] = None
    token: Optional[str] = None

    @root_validator
    def validate_credentials(cls, values):
        username, password = values.get("username"), values.get("password")
        token = values.get("token")
        client_id, client_secret = values.get("client_id"), values.get("client_secret")
        oidc_provider_url = values.get("oidc_provider_url")

        if username and not password:
            raise ValueError("Password is required if username is provided")
        if password and not username:
            raise ValueError("Username is required if password is provided")

        if (username and password) and token:
            raise ValueError("Cannot provide both username and password or token")

        if (client_id and client_secret) and token:
            raise ValueError("Cannot provide both client_id and client_secret or token")

        if (client_id and client_secret) and not oidc_provider_url:
            raise ValueError(
                "Cannot provide both client_id and client_secret without oidc_provider_url"
            )
        if (username and password) and client_id:
            raise ValueError("Cannot provide both username and password and client_id")

        return values


class FHIRServerCreate(FHIRServerBase):
    pass


class FHIRServerUpdate(FHIRServerBase):
    pass


class FHIRServer(FHIRServerBase):
    id: Any
    created_at: datetime
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True


class ServerStatistics(BaseModel):
    """
    Statistics for a FHIR Server
    """

    summary: ServerSummary
    created_at: datetime
    figure: Optional[dict] = None



import os

import pytest

from station.common.clients.central import CentralApiClient


@pytest.fixture
def central_robot_credentials():
    api_url = os.getenv("CENTRAL_API_URL")
    robot_id = os.getenv("STATION_ROBOT_ID")
    robot_secret = os.getenv("STATION_ROBOT_SECRET")

    return {
        "api_url": api_url,
        "robot_id": robot_id,
        "robot_secret": robot_secret,
    }


def test_get_registry_credentials(central_robot_credentials):
    station_id = os.getenv("STATION_ID")

    client = CentralApiClient(**central_robot_credentials)

    registry_credentials = client.get_registry_credentials(station_id)
    print(registry_credentials)
    assert registry_credentials
    assert registry_credentials.address
    assert registry_credentials.user
    assert registry_credentials.password
    assert registry_credentials.project



from typing import List

from fastapi import APIRouter, Depends, File, HTTPException, UploadFile
from fastapi.responses import StreamingResponse
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.config import clients
from station.app.crud.crud_local_train import local_train
from station.app.schemas import local_trains
from station.app.schemas.datasets import MinioFile
from station.common.constants import DataDirectories
from station.trains.local.airflow import run_local_train

router = APIRouter()


@router.post("", response_model=local_trains.LocalTrain)
def create_local_train(
    create_msg: local_trains.LocalTrainCreate,
    db: Session = Depends(dependencies.get_db),
):
    """
    creae a database entry for a new train with preset names from the create_msg

    @param create_msg: information about the new train
    @param db: reference to the postgres database
    @return:
    """
    train = local_train.create(db, obj_in=create_msg)
    return train


@router.get("/{train_id}", response_model=local_trains.LocalTrain)
def get_local_train(train_id: str, db: Session = Depends(dependencies.get_db)):
    train = local_train.get(db, train_id)
    if not train:
        raise HTTPException(status_code=404, detail=f"Train ({train_id}) not found")
    return train


@router.get("", response_model=List[local_trains.LocalTrain])
def get_local_trains(
    db: Session = Depends(dependencies.get_db), skip: int = 0, limit: int = 100
):
    trains = local_train.get_multi(db, skip=skip, limit=limit)

    return trains


@router.put("/{train_id}", response_model=local_trains.LocalTrain)
def update_local_train(
    train_id: str,
    update_msg: local_trains.LocalTrainUpdate,
    db: Session = Depends(dependencies.get_db),
):
    train = local_train.get(db, train_id)
    if not train:
        raise HTTPException(status_code=404, detail=f"Train ({train_id}) not found")

    train = local_train.update(db, db_obj=train, obj_in=update_msg)
    return train


@router.delete("/{train_id}", response_model=local_trains.LocalTrain)
def delete_local_train(train_id: str, db: Session = Depends(dependencies.get_db)):
    train = local_train.get(db, train_id)
    if not train:
        raise HTTPException(status_code=404, detail=f"Train ({train_id}) not found")

    train = local_train.remove(db, id=train_id)
    return train


@router.post("/{train_id}/run", response_model=local_trains.LocalTrainExecution)
async def trigger_train_execution(
    train_id: str,
    run_config: local_trains.LocalTrainRunConfig,
    db: Session = Depends(dependencies.get_db),
):
    train = local_train.get(db, train_id)
    if not train:
        raise HTTPException(status_code=404, detail=f"Train ({train_id}) not found")

    execution = run_local_train(
        db=db,
        train_id=train_id,
        config_id=run_config.config_id,
        dataset_id=run_config.dataset_id,
    )

    return execution


@router.post("/{train_id}/files")
async def upload_train_files(
    train_id: str,
    files: List[UploadFile] = File(description="Multiple files as UploadFile"),
    db: Session = Depends(dependencies.get_db),
) -> List[MinioFile]:
    db_train = local_train.get(db, train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Local train ({train_id}) not found."
        )
    if not files:
        raise HTTPException(status_code=400, detail="No files provided.")
    for file in files:
        if not file.filename:
            raise HTTPException(status_code=400, detail="No filename provided.")

    state = db_train.state

    if (
        state.configuration_state
        != local_trains.LocalTrainConfigurationStep.image_configured.value
    ):
        raise HTTPException(
            status_code=400,
            detail="Train image is not configured. Select an image first before up"
            "loading files.",
        )

    resp = await clients.minio.save_local_train_files(db_train.id, files)

    state.configuration_status = (
        local_trains.LocalTrainConfigurationStep.files_uploaded.value
    )
    db.commit()
    return resp


@router.get("/{train_id}/files", response_model=List[MinioFile])
async def get_train_files(
    train_id: str, file_name: str = None, db: Session = Depends(dependencies.get_db)
):
    db_train = local_train.get(db, train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Local train ({train_id}) not found."
        )

    items = clients.minio.get_minio_dir_items(
        DataDirectories.LOCAL_TRAINS.value, str(db_train.id)
    )
    if file_name:
        pass
    return items


@router.get("/{train_id}/archive", response_class=StreamingResponse)
async def get_train_archive(train_id: str, db: Session = Depends(dependencies.get_db)):
    db_train = local_train.get(db, train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Local train ({train_id}) not found."
        )
    resp = clients.minio.get_local_train_archive(str(db_train.id))
    return StreamingResponse(content=resp, media_type="application/octet-stream")



from typing import List

from station.app.models import local_trains as lt_models
from station.app.schemas import local_trains as lt_schema
from station.app.schemas.datasets import MinioFile


def update_configuration_status(
    local_train: lt_models.LocalTrain, files: List[MinioFile] = None
) -> str:
    """
    Update the configuration status of a local train
    """

    state = local_train.state.configuration_state

    if local_train.master_image_id or local_train.custom_image:
        state = lt_schema.LocalTrainConfigurationStep.image_configured.value
    if files:
        state = lt_schema.LocalTrainConfigurationStep.files_uploaded.value
    if local_train.entrypoint and local_train.master_image_id:
        state = lt_schema.LocalTrainConfigurationStep.image_configured.value
    if local_train.dataset_id:
        state = lt_schema.LocalTrainConfigurationStep.dataset_selected.value

    return state



import os
from enum import Enum

import click

from station.common.constants import DataDirectories, PHTDirectories, ServiceNames


def check_create_pht_dirs(path):
    """
    Check that the directory structure exists, if it does prompt the user for update/overwrite, otherwise create it.
    """
    # check that pht directories exist
    pht_dir_check = _check_dirs_from_enum(path, PHTDirectories)
    service_dir_check = _check_dirs_from_enum(
        os.path.join(path, PHTDirectories.SERVICE_DATA_DIR.value), ServiceNames
    )

    if pht_dir_check and service_dir_check:
        click.confirm(
            "Previous installation found. Do you want to overwrite it?", abort=True
        )
        create_pht_dirs(path)
    else:
        create_pht_dirs(path)


def create_pht_dirs(path):
    """
    Create the pht directory structure in the given directory.
    """
    if not os.path.exists(path):
        os.makedirs(path, mode=0o777)

    # check that pht directories ex
    _make_dirs_from_enum(path, PHTDirectories)
    # create subdirectories for storing service data
    service_path = os.path.join(path, PHTDirectories.SERVICE_DATA_DIR.value)
    _make_dirs_from_enum(service_path, ServiceNames)
    # create subdirectories for storing station data (mount point for minio data)
    data_path = os.path.join(path, PHTDirectories.STATION_DATA_DIR.value)
    _make_dirs_from_enum(data_path, DataDirectories)


def _make_dirs_from_enum(path, dir_enum: Enum):
    for dir in dir_enum:
        dir_path = os.path.join(path, dir.value)
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path, mode=0o777)


def _check_dirs_from_enum(path, dir_enum: Enum):
    for dir in dir_enum:
        dir_path = os.path.join(path, dir.value)
        if not os.path.isdir(dir_path):
            return False
    return True



import uuid
from datetime import datetime

from sqlalchemy import Boolean, Column, DateTime, String
from sqlalchemy.dialects.postgresql import UUID

from station.app.db.base_class import Base


class FHIRServer(Base):
    __tablename__ = "fhir_servers"
    id = Column(UUID(as_uuid=True), primary_key=True, index=True, default=uuid.uuid4)
    api_address = Column(String)
    name = Column(String, nullable=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, nullable=True)
    username = Column(String, nullable=True)
    password = Column(String, nullable=True)
    client_id = Column(String, nullable=True)
    client_secret = Column(String, nullable=True)
    oidc_provider_url = Column(String, nullable=True)
    token = Column(String)
    active = Column(Boolean, default=True)
    type = Column(String, nullable=True)
    proposal_id = Column(String, nullable=True)
    summary = Column(String, nullable=True)



from typing import List, Optional

from pydantic import BaseModel

from station.app.schemas.docker_trains import DockerTrain
from station.app.schemas.trains import Train


class Trains(BaseModel):
    docker_trains: Optional[List[DockerTrain]]
    federated_trains: Optional[List[Train]]

    class Config:
        orm_mode = True


class TrainOverViewResponse(BaseModel):
    active: Optional[Trains] = None
    available: Optional[Trains] = None



import tarfile
from io import BytesIO
from typing import List, Union

from train_lib.docker_util.docker_ops import add_archive

from docker.models.images import Image


def make_docker_file(
    master_image: str,
    entrypoint_file: str,
    command: str,
    command_args: Union[List[str], str] = None,
) -> BytesIO:
    """
    Make an in memory docker file for a local train
    Args:
        master_image: identifier of the base image to build the train on
        entrypoint_file: which file to use as the entrypoint
        command:
        command_args:

    Returns:

    """

    print(master_image, entrypoint_file, command, command_args)
    docker_from = f"FROM {master_image}\n"
    directory_setup = (
        "RUN mkdir /opt/train && mkdir /opt/results && chmod -R +x /opt/train \n"
    )

    if isinstance(command_args, str):
        command_args = command_args.split(" ")
    if command_args:
        command_args = [f'"{arg}"' for arg in command_args]
        str_args = ", ".join(command_args) + ", "
    else:
        str_args = ""

    docker_command = f'CMD ["{command}", {str_args}"/opt/train/{entrypoint_file}"]\n'

    docker_file = docker_from + directory_setup + docker_command

    return BytesIO(docker_file.encode("utf-8"))


def add_train_files(train_id: str, image: Image, files):
    archive = _make_archive(files)
    print(image.id)
    add_archive(f"{train_id}:latest", archive, path="/opt/train")


def _make_archive(files: list) -> BytesIO:
    archive = BytesIO()
    tar = tarfile.open(fileobj=archive, mode="w")

    tar.close()
    # reset the archive
    archive.seek(0)

    return archive



from enum import Enum

CERTS_REGEX = r"https\.certs\[([0-9]*)\]"


class ApplicationEnvironment(str, Enum):
    DEVELOPMENT = "development"
    PRODUCTION = "production"


class DockerVolumes(Enum):
    POSTGRES = "pg_pht_station"


class DockerNetworks(Enum):
    STATION = "pht-station"


class PHTImages(Enum):
    API = "ghcr.io/pht-medic/station-api"
    UI = "ghcr.io/pht-medic/station-ui"
    AUTH = "ghcr.io/authup/authup"
    AIRFLOW = "ghcr.io/pht-medic/airflow"


class ServiceImages(Enum):
    MINIO = "minio/minio:RELEASE.2022-11-11T03-44-20Z"
    POSTGRES = "postgres:14"
    REDIS = "redislabs/rejson:latest"
    TRAEFIK = "traefik:v2.8"
    BLAZE = "samply/blaze:latest"


class DefaultValues(Enum):
    """
    Default values for the station configuration
    """

    FERNET_KEY = "your_fernet_key"
    ADMIN = "admin"
    ADMIN_PASSWORD = "password"
    PRIVATE_KEY = "/path/to/private_key.pem"
    STATION_DOMAIN = "example-station.com"
    CERT = "example-cert.pem"
    KEY = "example-key.pem"
    DOMAIN = "station.localhost"
    ROBOT_ID = "central-robot-id"
    ROBOT_SECRET = "central-robot-secret"
    HTTP_PORT = 80
    HTTPS_PORT = 443


class PHTDirectories(Enum):
    SERVICE_DATA_DIR = "service_data"
    SERVICE_LOG_DIR = "logs"
    CONFIG_DIR = "configs"
    CERTS_DIR = "certs"
    STATION_DATA_DIR = "station_data"
    SETUP_SCRIPT_DIR = "setup_scripts"


class DataDirectories(str, Enum):
    MODELS = "models"
    TRAINS = "trains"
    DATASETS = "datasets"
    LOCAL_TRAINS = "localtrains"


class ServiceNames(Enum):
    AUTH = "auth"
    POSTGRES = "postgres"
    REDIS = "redis"
    MINIO = "minio"


class Icons(Enum):
    CHECKMARK = "✓"
    CROSS = "❌"
    WARNING = "⚠"



import os

import pytest
from dotenv import find_dotenv, load_dotenv
from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


@pytest.fixture
def train_id():
    load_dotenv(find_dotenv())
    return "testTrain"


@pytest.fixture(scope="session")
def docker_train_config():
    config = {
        "name": "test_config",
        "airflow_config": {
            "env": [{"key": "FHIR_ADDRESS", "value": "test_address"}],
            "volumes": [
                {
                    "host_path": "path/on/host",
                    "container_path": "path/in/container",
                    "mode": "ro",
                }
            ],
            "repository": "example/repository",
            "tag": "latest",
        },
        "auto_execute": True,
    }

    return config


@pytest.fixture(scope="session")
def config_id(docker_train_config):
    response = client.post("/api/trains/docker/config", json=docker_train_config)
    return response.json()["id"]


@pytest.fixture(scope="session", autouse=True)
def train_list_length():
    response = client.get("/api/trains/docker")
    length = len(response.json())
    return length


def test_config_create(docker_train_config, config_id):
    response = client.get(f"/api/trains/docker/config/{config_id}")
    assert response.status_code == 200, response.text

    assert response.json()["name"] == docker_train_config["name"]
    assert response.json()["auto_execute"]


def test_docker_train_create(train_id):
    response = client.get("/api/trains/docker/configs/all")
    print(response.json())
    response = client.post("/api/trains/docker", json={"train_id": train_id})

    assert response.status_code == 200, response.text

    json_response = response.json()
    assert json_response["train_id"] == train_id
    print(json_response)
    assert json_response["state"]["num_executions"] == 0
    assert json_response["state"]["status"] == "inactive"


def test_docker_train_create_fails(train_id):
    response = client.post("/api/trains/docker", json={"train_id": train_id})

    assert response.status_code == 400, response.text


def test_get_train_by_id(train_id):
    response = client.get(f"/api/trains/docker/{train_id}")
    assert response.status_code == 200, response.text


def test_get_train_by_id_fails():
    response = client.get("/api/trains/docker/notthere")
    assert response.status_code == 404, response.text


def test_list_docker_trains(train_list_length):
    response = client.get("/api/trains/docker")

    assert response.status_code == 200, response.text
    print(response.json())

    assert len(response.json()) == train_list_length + 1


def test_docker_train_config_create_fails(docker_train_config):
    response = client.post("/api/trains/docker/config", json=docker_train_config)
    assert response.status_code == 400


def test_get_docker_train_configs():
    response = client.get("/api/trains/docker/configs/all")
    assert response.status_code == 200, response.text
    assert len(response.json()) >= 1


def test_get_docker_train_config_by_id(config_id):
    response = client.get("/api/trains/docker/config/1")
    assert response.status_code == 200, response.text


def test_get_docker_train_config_by_id_fails():
    response = client.get("/api/trains/docker/config/234")
    assert response.status_code == 404, response.text


def test_update_docker_train_config(docker_train_config, config_id):
    docker_train_config["name"] = "updated name"
    response = client.put(
        f"/api/trains/docker/config/{config_id}", json=docker_train_config
    )

    assert response.status_code == 200, response.text
    response = client.get(f"/api/trains/docker/config/{config_id}")

    assert response.json()["name"] == "updated name"


def test_update_docker_train_config_fails(docker_train_config):
    docker_train_config["name"] = "updated name"
    response = client.put("/api/trains/docker/config/234", json=docker_train_config)
    assert response.status_code == 404, response.text


def test_assign_docker_train_config(train_id, config_id):
    response = client.post(f"/api/trains/docker/{train_id}/config/{config_id}")
    assert response.status_code == 200, response.text
    response = client.get(f"/api/trains/docker/{train_id}")

    assert response.json()["config_id"] == config_id

    # test non existing config error
    response = client.post(f"/api/trains/docker/{train_id}/config/321")

    assert response.status_code == 404

    # test non existing train error

    response = client.post("/api/trains/docker/no_train/config/1")
    assert response.status_code == 404


def test_get_config_for_train(train_id):
    response = client.get(f"/api/trains/docker/{train_id}/config")
    assert response.status_code == 200, response.text
    assert len(response.json()["trains"]) == 1
    print(response.json())

    new_train_id = "no_config_train"
    response = client.post("/api/trains/docker", json={"train_id": new_train_id})
    assert response.status_code == 200
    response = client.get(f"/api/trains/docker/{new_train_id}/config")

    assert response.status_code == 404


def test_create_train_with_config(docker_train_config, config_id):
    # assign existing config
    response = client.post(
        "/api/trains/docker",
        json={"train_id": "with_existing_config", "config": config_id},
    )
    assert response.status_code == 200, response.text
    config_response = client.get("/api/trains/docker/with_existing_config/config")
    assert config_response.json()["name"] == "updated name"
    assert len(config_response.json()["trains"]) == 2

    # fails with unknown config id
    response = client.post(
        "/api/trains/docker", json={"train_id": "config does not exist", "config": 3213}
    )
    assert response.status_code == 404, response.text

    new_config = {
        "name": "new config",
        "airflow_config": {
            "env": [{"key": "FHIR_ADDRESS", "value": "test_address"}],
            "volumes": [
                {
                    "host_path": "path/on/host",
                    "container_path": "path/in/container",
                    "mode": "ro",
                }
            ],
        },
        "auto_execute": False,
    }

    response = client.post(
        "/api/trains/docker",
        json={
            "train_id": "with_new_config",
            "name": "train with new config",
            "config": new_config,
        },
    )
    assert response.status_code == 200
    print(response.json())

    config_response = client.get("/api/trains/docker/with_new_config/config")
    assert config_response.json()["name"] == "new config"


def test_get_train_state():
    train_id = "test_train_state"
    response = client.post(
        "/api/trains/docker",
        json={
            "train_id": train_id,
        },
    )
    assert response.status_code == 200, response.text

    response = client.get(f"/api/trains/docker/{train_id}/state")
    assert response.status_code == 200, response.text
    assert response.json()["status"] == "inactive"


def test_update_train_state(train_id):
    response = client.put(
        f"/api/trains/docker/{train_id}/state",
        json={"status": "active", "num_executions": 1},
    )
    assert response.status_code == 200, response.text
    assert response.json()["status"] == "active"
    assert response.json()["num_executions"] == 1


def test_synchronize_database():
    if os.getenv("ENVIRONMENT") == "testing":
        response_nostation = client.get("/api/trains/docker/sync")
        assert response_nostation.status_code == 200, response_nostation.text

        response = client.get("/api/trains/docker/sync/?station_id=1")
        if os.getenv("STATION_ID") == 1:
            assert len(response.json()) == 0
        assert response.status_code == 200, response.text


def test_synchronize_database_fails():
    if os.getenv("ENVIRONMENT") == "testing":
        response = client.get("/api/trains/docker/sync/?station_id=123")
        assert response.status_code == 404, response.text


def test_run_docker_train(train_id, docker_train_config, config_id):
    if os.getenv("ENVIRONMENT") == "testing":
        old_state = client.get(f"/api/trains/docker/{train_id}/state")

        # run with given config id
        response = client.post(
            f"/api/trains/docker/{train_id}/run", json={"config_id": 1}
        )
        assert response.json()["airflow_dag_run"]
        assert response.json()["config"] == 1
        assert response.status_code == 200, response.text

        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        assert old_state.json() != state_response.json()
        assert (
            old_state.json()["num_executions"] + 1
            == state_response.json()["num_executions"]
        )

        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert (
            execution_response.json()[-1]["airflow_dag_run"]
            == response.json()["airflow_dag_run"]
        )

        # run with default config
        response = client.post(
            f"/api/trains/docker/{train_id}/run", json={"config_id": "default"}
        )
        assert response.json()["airflow_dag_run"]
        assert response.json()["config"] is None
        assert response.status_code == 200, response.text

        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        assert (
            old_state.json()["num_executions"] + 2
            == state_response.json()["num_executions"]
        )

        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert (
            execution_response.json()[-1]["airflow_dag_run"]
            == response.json()["airflow_dag_run"]
        )

        # run with config assigned to train
        response = client.post(f"/api/trains/docker/{train_id}/run")
        assert response.json()["airflow_dag_run"]
        assert response.json()["config"] == config_id
        assert response.status_code == 200, response.text

        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        assert (
            old_state.json()["num_executions"] + 3
            == state_response.json()["num_executions"]
        )

        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert (
            execution_response.json()[-1]["airflow_dag_run"]
            == response.json()["airflow_dag_run"]
        )


def test_run_docker_train_fails(train_id, docker_train_config, config_id):
    old_state = client.get(f"/api/trains/docker/{train_id}/state")
    old_executions = client.get(f"/api/trains/docker/{train_id}/executions")

    if os.getenv("ENVIRONMENT") == "testing":

        # no config with id given
        response = client.post(
            f"/api/trains/docker/{train_id}/run", json={"config_id": 3213}
        )
        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert old_state.json() == state_response.json()
        assert old_executions.json() == execution_response.json()
        assert response.status_code == 400, response.text

        # no tag and no repository given
        specific_id = config_id + 1
        response = client.post(
            f"/api/trains/docker/{train_id}/run", json={"config_id": specific_id}
        )
        response_conf = client.get(f"/api/trains/docker/config/{specific_id}")
        print(response_conf.json())
        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert old_state.json() == state_response.json()
        assert old_executions.json() == execution_response.json()
        assert response.status_code == 400, response.text

        # train not defined
        response = client.post("/api/trains/docker/no_train/run", json={"config_id": 1})
        assert response.status_code == 404, response.text

    else:
        response = client.post(
            f"/api/trains/docker/{train_id}/run", json={"config_id": "default"}
        )
        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert old_state.json() == state_response.json()
        assert old_executions.json() == execution_response.json()
        assert response.status_code == 503, response.text

        response = client.post(f"/api/trains/docker/{train_id}/run")
        state_response = client.get(f"/api/trains/docker/{train_id}/state")
        execution_response = client.get(f"/api/trains/docker/{train_id}/executions")
        assert old_state.json() == state_response.json()
        assert old_executions.json() == execution_response.json()
        assert response.status_code == 503, response.text



from typing import List

from fastapi import APIRouter, Depends
from sqlalchemy.orm import Session

from station.app.api import dependencies
from station.app.crud.crud_fhir_servers import fhir_servers
from station.app.fhir.server import get_server_statistics
from station.app.schemas.fhir import (
    FHIRServer,
    FHIRServerCreate,
    FHIRServerUpdate,
    ServerStatistics,
)

router = APIRouter()


# todo error handling


@router.post("", response_model=FHIRServer, status_code=201)
def add_fhir_server(
    fhir_server_in: FHIRServerCreate, db: Session = Depends(dependencies.get_db)
):
    db_fhir_server = fhir_servers.create(db=db, obj_in=fhir_server_in)
    return db_fhir_server


@router.get("", response_model=List[FHIRServer])
def get_fhir_servers(
    limit: int = 100, skip: int = 0, db: Session = Depends(dependencies.get_db)
):
    db_fhir_servers = fhir_servers.get_multi(db=db, skip=skip, limit=limit)
    return db_fhir_servers


@router.put("/{server_id}", response_model=FHIRServer)
def update_fhir_server(
    server_id: str,
    update_in: FHIRServerUpdate,
    db: Session = Depends(dependencies.get_db),
):
    db_fhir_server = fhir_servers.get(db, id=server_id)
    db_fhir_server = fhir_servers.update(db=db, db_obj=db_fhir_server, obj_in=update_in)
    return db_fhir_server


@router.delete("/{server_id}", status_code=202, response_model=FHIRServer)
def delete_fhir_server(server_id: str, db: Session = Depends(dependencies.get_db)):
    deleted_server = fhir_servers.remove(db=db, id=server_id)
    return deleted_server


@router.get("/{server_id}", response_model=FHIRServer)
def get_fhir_server(server_id: str, db: Session = Depends(dependencies.get_db)):
    db_fhir_server = fhir_servers.get(db=db, id=server_id)
    return db_fhir_server


@router.get("/{server_id}/stats", response_model=ServerStatistics)
def fhir_server_summary(
    server_id: str, refresh: bool = False, db: Session = Depends(dependencies.get_db)
):
    server_stats = get_server_statistics(db, fhir_server_id=server_id, refresh=refresh)
    return server_stats



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, validator


class DBSchema(BaseModel):
    class Config:
        orm_mode = True


class DockerTrainState(DBSchema):
    num_executions: Optional[int] = 0
    status: Optional[str] = "inactive"
    last_execution: Optional[datetime] = None
    central_status: Optional[str] = None


class AirflowEnvironmentVariable(BaseModel):
    key: str
    value: str


class DockerVolume(BaseModel):
    host_path: str
    container_path: str
    mode: Optional[str] = "ro"


class AirflowConfig(BaseModel):
    env: Optional[List[AirflowEnvironmentVariable]] = None
    volumes: Optional[List[DockerVolume]] = None


class DockerTrainAirflowConfig(AirflowConfig):
    repository: Optional[str] = None
    tag: Optional[str] = None


class DockerTrainConfigBase(DBSchema):
    name: str
    airflow_config: Optional[DockerTrainAirflowConfig] = None
    cpu_requirements: Optional[Dict[str, Any]] = None
    gpu_requirements: Optional[Dict[str, Any]] = None
    auto_execute: Optional[bool] = None


class DockerTrainMinimal(DBSchema):
    train_id: Optional[str] = None


class DockerTrainConfig(DockerTrainConfigBase):
    id: int
    created_at: datetime
    updated_at: Optional[datetime] = None
    trains: Optional[List[DockerTrainMinimal]] = None

    # concatenates list of objects with train_ids to one list of train_ids
    @validator("trains")
    def train_list(cls, v):
        if v:
            train_list = []
            for train in v:
                train_list.append(train.train_id)
            return train_list
        else:
            return None


class DockerTrainExecution(DBSchema):
    config_id: Optional[Union[int, str]] = "default"
    dataset_id: Optional[Union[int, str]] = None


class DockerTrainSavedExecution(DBSchema):
    start: datetime
    end: Optional[datetime] = None
    airflow_dag_run: Optional[str] = None
    config: Optional[int] = None
    dataset: Optional[str] = None
    train_id: Optional[str] = None


class DockerTrain(DBSchema):
    name: Optional[str] = None
    created_at: datetime
    updated_at: Optional[datetime] = None
    proposal: Optional[str] = None
    type: Optional[str] = None
    is_active: bool = False
    train_id: Optional[str] = None
    config_id: Optional[int] = None
    image_name: Optional[str] = None
    num_participants: Optional[int] = None
    state: Optional[DockerTrainState] = None
    executions: Optional[List[DockerTrainSavedExecution]] = None


class DockerTrainConfigCreate(DockerTrainConfigBase):
    pass


class DockerTrainConfigUpdate(DockerTrainConfigBase):
    pass


class DockerTrainCreate(BaseModel):
    train_id: str
    proposal_id: Optional[int] = None
    config: Optional[Union[DockerTrainConfigCreate, int]] = None


class DockerTrainUpdate(DockerTrainCreate):
    pass



import json
from datetime import datetime

import plotly.graph_objects as go
import plotly.io
from fhir_kindling import FhirServer
from fhir_kindling.fhir_server.server_responses import ServerSummary
from sqlalchemy.orm import Session

from station.app.crud.crud_fhir_servers import fhir_servers
from station.app.schemas.fhir import ServerStatistics
from station.app.settings import settings


def fhir_server_from_db(db: Session, fhir_server_id: str, server=None) -> FhirServer:
    if server:
        db_server = server
    else:
        db_server = fhir_servers.get(db, id=fhir_server_id)

    assert db_server
    # for each of the different options access and decrypt the sensitive values and initialize a server instance
    if db_server.username:
        password = settings.get_fernet().decrypt(db_server.password.encode()).decode()
        return FhirServer(
            api_address=db_server.api_address,
            username=db_server.username,
            password=password,
            fhir_server_type=db_server.type,
        )

    if db_server.token:
        token = settings.get_fernet().decrypt(db_server.token.encode()).decode()
        return FhirServer(
            api_address=db_server.api_address,
            token=token,
            fhir_server_type=db_server.type,
        )

    if db_server.client_id:
        client_secret = settings.get_fernet().decrypt(db_server.client_secret.encode())
        return FhirServer(
            api_address=db_server.api_address,
            fhir_server_type=db_server.type,
            client_id=db_server.client_id,
            client_secret=client_secret.decode("utf-8"),
            oidc_provider_url=db_server.oidc_provider_url,
        )


def get_server_statistics(
    db: Session, fhir_server_id: str, refresh: bool = False
) -> ServerStatistics:
    db_server = fhir_servers.get(db, id=fhir_server_id)

    if db_server.summary and not refresh:
        return ServerStatistics.parse_raw(db_server.summary)

    server = fhir_server_from_db(db, fhir_server_id, db_server)
    summary = server.summary()
    obj = summary_plot(summary)
    return ServerStatistics(created_at=datetime.now(), summary=summary, figure=obj)


def summary_plot(summary: ServerSummary) -> dict:
    # sort the resources by count
    resources = sorted(summary.resources, key=lambda x: x.count)
    resources, counts = zip(*[(r.resource, r.count) for r in resources if r.count > 0])
    fig = go.Figure(go.Bar(x=list(counts), y=list(resources), orientation="h"))

    fig_json = plotly.io.to_json(fig)
    obj = json.loads(fig_json)
    return obj



import os

import s3fs

from station.app.settings import settings


def get_s3_filesystem(
    minio_url: str = None, access_key: str = None, secret_key: str = None
) -> s3fs.S3FileSystem:
    """
    Returns filesystem object to access files saved in minio
    :param minio_url: Url to the minio server
    :param access_key: Minio access key
    :param secret_key: Minio secret key
    :return: S3 Filesystem
    """
    if minio_url is None:
        minio_host = settings.config.minio.host
        minio_port = settings.config.minio.port
        minio_url = minio_host + ":" + str(minio_port)

    assert minio_url is not None, "MINIO_URL is not set"

    if access_key is None:
        access_key = settings.config.minio.access_key

    if secret_key is None:
        secret_key = settings.config.minio.secret_key.get_secret_value()

    fs = s3fs.S3FileSystem(
        anon=False,
        key=access_key,
        secret=secret_key,
        use_ssl=True,
        client_kwargs={"endpoint_url": f"http://{minio_url}/"},
    )

    return fs


def get_file(path, storage_type):
    """
    Returns file independent of saved location (local or on minio server)
    :param path: Path to file
    :return: File
    """
    # access file on minio server
    if storage_type == "minio":
        fs = get_s3_filesystem()
        # get file
        try:
            file = fs.open(path)
            return file
        except Exception as e:
            print(e)
            raise FileNotFoundError
    elif storage_type == "local":
        try:
            path = os.path.join(settings.config.station_data_dir, path)
            file = open(path)
            return file
        except Exception as e:
            print(e)
            raise FileNotFoundError
    else:
        raise NotImplementedError



import os
from unittest.mock import patch

import pytest
from dotenv import find_dotenv, load_dotenv

from station.app.env import StationEnvironmentVariables
from station.app.settings import (
    AirflowSettings,
    AuthConfig,
    CentralUISettings,
    MinioSettings,
    RegistrySettings,
    Settings,
    StationConfig,
)


def test_settings_init_env_vars():
    load_dotenv(find_dotenv())
    # Test runtime environment variables
    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            StationEnvironmentVariables.AUTH_SERVER_HOST.value: "http://auth.example.com",
            StationEnvironmentVariables.AUTH_SERVER_PORT.value: "3010",
            StationEnvironmentVariables.AUTH_ROBOT_ID.value: "robot",
            StationEnvironmentVariables.AUTH_ROBOT_SECRET.value: "robot_secret",
            StationEnvironmentVariables.MINIO_HOST.value: "http://minio.example.com",
            StationEnvironmentVariables.MINIO_ACCESS_KEY.value: "minio_user",
            StationEnvironmentVariables.MINIO_SECRET_KEY.value: "minio_secret",
            StationEnvironmentVariables.STATION_DB.value: "postgres://db",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.environment == "development"
        assert settings.config.auth.host == "http://auth.example.com"
        assert settings.config.auth.port == 3010
        assert settings.config.auth.robot_id == "robot"

    with patch.dict(os.environ, {"ENVIRONMENT": "development"}):
        settings = Settings()
        settings.setup()
        assert settings.config.environment == "development"

    with pytest.raises(ValueError):
        with patch.dict(os.environ, {"ENVIRONMENT": "fails"}):
            settings = Settings()
            settings.setup()

    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            "STATION_ID": "test_station_id",
            "FERNET_KEY": "test_fernet_key",
            "STATION_API_HOST": "0.0.0.0",
            "STATION_API_PORT": "8082",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.station_id == "test_station_id"
        assert settings.config.environment == "development"
        assert settings.config.fernet_key == "test_fernet_key"
        assert settings.config.host == "0.0.0.0"
        assert settings.config.port == 8082

    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            "STATION_ID": "test_station_id",
            "FERNET_KEY": "",
            "STATION_API_HOST": "0.0.0.0",
            "STATION_API_PORT": "8082",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.station_id == "test_station_id"
        assert settings.config.environment == "development"
        assert settings.config.fernet_key
        assert settings.config.host == "0.0.0.0"
        assert settings.config.port == 8082

    with pytest.raises(ValueError):
        with patch.dict(
            os.environ,
            {
                "ENVIRONMENT": "development",
                "STATION_ID": "",
                "FERNET_KEY": "",
                "STATION_API_HOST": "0.0.0.0",
                "STATION_API_PORT": "8082",
            },
        ):
            settings = Settings()
            settings.setup()

    with pytest.raises(ValueError):
        with patch.dict(
            os.environ,
            {
                "ENVIRONMENT": "production",
                StationEnvironmentVariables.AUTH_SERVER_HOST.value: "",
                StationEnvironmentVariables.AUTH_SERVER_PORT.value: "",
                StationEnvironmentVariables.AUTH_ROBOT_ID.value: "",
                StationEnvironmentVariables.AUTH_ROBOT_SECRET.value: "",
            },
        ):
            settings = Settings()
            settings.setup()
    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            StationEnvironmentVariables.AUTH_SERVER_HOST.value: "",
            StationEnvironmentVariables.AUTH_SERVER_PORT.value: "",
            StationEnvironmentVariables.AUTH_ROBOT_ID.value: "",
            StationEnvironmentVariables.AUTH_ROBOT_SECRET.value: "",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.environment == "development"
        assert settings.config.auth is None

    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            StationEnvironmentVariables.REGISTRY_URL.value: "http://registry.example.com",
            StationEnvironmentVariables.REGISTRY_USER.value: "test",
            StationEnvironmentVariables.REGISTRY_PW.value: "test",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.environment == "development"
        assert settings.config.registry.address == "http://registry.example.com"
        assert settings.config.registry.user == "test"
        assert settings.config.registry.password.get_secret_value() == "test"

    with patch.dict(
        os.environ,
        {
            "ENVIRONMENT": "development",
            StationEnvironmentVariables.MINIO_HOST.value: "http://registry.example.com",
            StationEnvironmentVariables.MINIO_ACCESS_KEY.value: "test",
            StationEnvironmentVariables.MINIO_SECRET_KEY.value: "test",
        },
    ):
        settings = Settings()
        settings.setup()
        assert settings.config.environment == "development"
        assert settings.config.minio.host == "http://registry.example.com"
        assert settings.config.minio.secret_key.get_secret_value() == "test"
        assert settings.config.minio.access_key == "test"


def test_config_file():
    load_dotenv(find_dotenv())
    settings = Settings()
    config = settings.setup()
    # write a working config
    config.station_id = "your_station_id"
    config.environment = "production"
    config.fernet_key = "your_fernet_key"
    config.host = "127.0.0.1"
    config.port = 8001
    config.auth = AuthConfig(
        host="127.0.0.1",
        port=3010,
        robot_id="your_robot_id",
        robot_secret="your_robot_secret",
    )
    config.registry = RegistrySettings(
        address="http://registry.example.com",
        user="test",
        password="test",
    )
    config.minio = MinioSettings(
        host="minio",
        port=9000,
        access_key="minio_access_key",
        secret_key="minio_secret_key",
    )
    config.central_ui = CentralUISettings(
        api_url="http://central-ui.example.com",
        client_id="central_ui_client_id",
        client_secret="central_ui_client_secret",
    )
    config = StationConfig(**settings.config.dict())

    config.to_file("station_config.yml")

    file_config = StationConfig.from_file("station_config.yml")
    assert file_config.station_id == "your_station_id"
    assert file_config.registry.user == "test"


def test_airflow_settings():

    airflow_settings = AirflowSettings(
        host="airflow",
        port=8080,
        user="airflow",
        password="airflow",
    )
    assert airflow_settings.host == "airflow"

    assert airflow_settings.api_url == "http://airflow:8080/api/v1/"

    airflow_settings = AirflowSettings(
        host="http://airflow.example.com",
        user="airflow",
        password="airflow",
    )

    assert airflow_settings.api_url == "http://airflow.example.com/api/v1/"

    airflow_settings = AirflowSettings(
        host="http://airflow.example.com/api/v1/",
        user="airflow",
        password="airflow",
    )

    assert airflow_settings.api_url == "http://airflow.example.com/api/v1/"


# def test_settings():
#     settings = Settings(config_path="station_config.yml")
#     settings.setup()
#     assert settings.config
#     os.remove("station_config.yml")



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class TorchModelBase(BaseModel):
    model_id: Optional[str] = None


class TorchModelCreate(TorchModelBase):
    pass


class TorchModelUpdate(TorchModelBase):
    pass


class TorchModelCheckPoint(BaseModel):
    id: int
    created_at: datetime
    updated_at: Optional[datetime]
    checkpoint: Optional[bytes]

    class Config:
        orm_mode = True


class TorchModel(TorchModelBase):
    id: int
    created_at: datetime
    updated_at: Optional[datetime]

    class Config:
        orm_mode = True


class DLModelBase(BaseModel):
    model_type: str
    model_name: str
    model_src: str
    train_id: Optional[str] = None


class DLModelCreate(DLModelBase):
    pass


class DLModelUpdate(DLModelBase):
    pass


class DLModel(DLModelBase):
    id: int
    model_id: Optional[str] = None
    created_at: datetime
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True



from typing import Union

from sqlalchemy.orm import Session

from station.app.crud import docker_trains
from station.common.clients.harbor_client import HarborClient


def sync_db_with_registry(db: Session, station_id: Union[str, int] = None):
    """
    Sync the stations local docker train database with the harbor project associated with the station

    :param db: database handle
    :param station_id: Optional Parameter identifying the station, loads .env var "STATION_ID" if not given
    :return:
    """
    harbor_client = HarborClient()
    harbor_repos = harbor_client.get_artifacts_for_station(station_id)

    for repo in harbor_repos:
        train_id = repo["name"].split("/")[-1]
        docker_trains.add_if_not_exists(
            db=db, train_id=train_id, created_at=repo["creation_time"]
        )



from io import BytesIO

import requests

from station.app.schemas.local_trains import (
    LocalTrain,
    LocalTrainCreate,
    LocalTrainUpdate,
)
from station.common.clients.resource_client import ResourceClient


class LocalTrainClient(ResourceClient[LocalTrain, LocalTrainCreate, LocalTrainUpdate]):
    def download_train_archive(self, train_id: str) -> BytesIO:
        url = f"{self.base_url}/{self.resource_name}/{train_id}/archive"
        with requests.get(url, headers=self._client.headers, stream=True) as r:
            r.raise_for_status()
            file_obj = BytesIO()
            for chunk in r.iter_content():
                file_obj.write(chunk)
            file_obj.seek(0)
        return file_obj

    def post_failure_notification(
        self,
        train_id: str,
        message: str,
    ) -> None:
        url = f"{self.base_url}/notifications"
        payload = {
            "message": message,
            "topic": "local-trains",
            "title": f"Local Train {train_id} failed",
        }
        with requests.post(url, headers=self._client.headers, json=payload) as r:
            r.raise_for_status()

    def update_train_status(self, train_id: str, status: str):
        url = f"{self.base_url}/{self.resource_name}/{train_id}"
        payload = {"status": status}
        with requests.put(url, headers=self._client.headers, json=payload) as r:
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                print(e)
                print(r.text)
                raise e



import os
from datetime import datetime
from typing import Any, Tuple, Union

from fastapi import HTTPException
from loguru import logger
from sqlalchemy.orm import Session

from station.app.config import clients, settings
from station.app.crud.crud_datasets import datasets
from station.app.crud.crud_docker_trains import docker_trains
from station.app.crud.crud_train_configs import docker_train_config
from station.app.models import docker_trains as dtm
from station.app.schemas import docker_trains as dts


def run_train(
    db: Session, train_id: Any, execution_params: dts.DockerTrainExecution
) -> dts.DockerTrainSavedExecution:
    """
    Execute a PHT 1.0 docker train using a configured airflow instance

    :param db: database session
    :param train_id: identifier of the train
    :param execution_params: given config_id or config_json can be used for running train
    :return:
    """
    # Extract the train from the database
    db_train = docker_trains.get_by_train_id(db, train_id)
    if not db_train:
        raise HTTPException(
            status_code=404, detail=f"Train with id '{train_id}' not found."
        )

    # Use default config if there is no config defined.
    if execution_params is None:
        config_id = db_train.config_id
        if not config_id:
            config_id = "default"
        execution_params = dts.DockerTrainExecution(config_id=config_id)

    config_id, config_dict = validate_run_config(db, train_id, execution_params)
    # process assigned data sets
    if execution_params.dataset_id:
        dataset = datasets.get(db, execution_params.dataset_id)
        process_dataset(config_dict, dataset)

    # Execute the train using the airflow rest api
    try:
        run_id = clients.airflow.trigger_dag("run_pht_train", config=config_dict)
        db_train = update_train_after_run(
            db, db_train, run_id, config_id, dataset_id=execution_params.dataset_id
        )
        last_execution = db_train.executions[-1]
        return last_execution
    except Exception as e:
        logger.error(
            f"Error while running train {train_id} with config {config_dict} \n {e}"
        )
        raise HTTPException(
            status_code=503,
            detail="No connection to the airflow client could be established.",
        )


def validate_run_config(
    db: Session,
    train_id: str,
    execution_params: dts.DockerTrainExecution,
    tag: str = None,
) -> Tuple[Union[int, str], dict]:
    """
    Validate the config used for the triggered run
    :param db: database session
    :param train_id: train id of the train to run
    :param execution_params: includes the config_id of the config to use or the specified config
    :param tag: optional tag of the image
    :return:

    """

    harbor_url = settings.config.registry.address
    project = settings.config.registry.project
    config = {
        "repository": f"{harbor_url}/{project}/{train_id}",
        "tag": "latest" if not tag else tag,
    }

    # Extract config by id if given
    if execution_params.config_id != "default":
        db_config = docker_train_config.get(db, execution_params.config_id)
        process_db_config(config, db_config)
        return db_config.id, config
    # Using the default config
    else:
        logger.info(f"Starting train {train_id} using default config")
        # Default config specifies only the identifier of the the train image and uses the latest tag
        return "default", config


def update_state(
    db: Session, db_train: dtm.DockerTrain, run_time: datetime
) -> dts.DockerTrainState:
    """
    Update the train state object of the train after starting an execution.
    Args:
        db: database session
        db_train: train object
        run_time: time when the run was triggered

    Returns:
        updated train state object

    """
    train_state = (
        db.query(dtm.DockerTrainState)
        .filter(dtm.DockerTrainState.train_id == db_train.id)
        .first()
    )
    if train_state:
        train_state.last_execution = run_time
        train_state.num_executions += 1
        train_state.status = "active"
    else:
        logger.info("No train state assigned.")
    db.add(train_state)
    db.commit()
    db.refresh(train_state)

    return train_state


def update_train_after_run(
    db: Session,
    db_train: dtm.DockerTrain,
    run_id: str,
    config_id: int,
    dataset_id: int = None,
) -> dts.DockerTrain:
    """
    Update train in the database and create a new execution object that stores the run configuration.

    Args:
        db: database session
        db_train: train object
        run_id: run id of the airflow DAG run
        config_id: id of the config used for the run
        dataset_id: id of the dataset used for the run

    Returns:
        updated train object
    """
    db_train.is_active = True
    run_time = datetime.now()
    db_train.updated_at = run_time

    # Update the train state
    update_state(db, db_train, run_time)

    # Create an execution
    execution = dtm.DockerTrainExecution(
        train_id=db_train.id,
        airflow_dag_run=run_id,
        config=config_id,
        dataset=dataset_id,
    )
    db.add(execution)
    db.commit()
    db.refresh(execution)

    db.commit()

    return db_train


def process_db_config(config_dict: dict, db_config: dtm.DockerTrainConfig) -> dict:
    """
    Update the config dictionary and with the values from db configuration
    :param config_dict: config dictionary
    :param db_config: db_config object
    :return:
    """

    if db_config.airflow_config:
        db_config = dts.DockerTrainConfig.from_orm(db_config)
        env_dict = {}
        for env in db_config.airflow_config.env:
            env_dict[env.key] = env.value
        config_dict["env"] = env_dict

        volume_dict = {}
        for volume in db_config.airflow_config.volumes:
            volume_dict[volume.host_path] = {
                "bind": volume.container_path,
                "mode": volume.mode,
            }
        config_dict["volumes"] = volume_dict
        return config_dict


def process_dataset(config_dict, dataset):
    """
    Update the config dictionary with the values from the dataset
    Args:
        config_dict:
        dataset:

    Returns:

    """
    mount_path = os.path.join(
        settings.config.station_data_dir, "datasets", str(dataset.id)
    )
    volumes = config_dict.get("volumes", {})
    volumes[mount_path] = {"bind": "/opt/train_data/", "mode": "ro"}
    config_dict["volumes"] = volumes



import os
import random
import string
from typing import Any

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import BaseModel


class GeneratorResult(BaseModel):
    """Base class for generator results"""

    loc: tuple
    value: Any


def password_generator() -> str:
    return "".join(
        [random.choice(string.ascii_letters + string.digits) for _ in range(32)]
    )


def generate_private_key(
    name: str, path: str = None, password: str = None
) -> list[GeneratorResult]:
    private_key = rsa.generate_private_key(
        65537, key_size=2048, backend=default_backend()
    )
    # encrypt key with password when given
    if password:
        encryption_algorithm = serialization.BestAvailableEncryption(password.encode())
    else:
        encryption_algorithm = serialization.NoEncryption()

    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=encryption_algorithm,
    )

    # name the private key file as pem
    if name.split(".")[-1] != "pem":
        name += ".pem"

    # if a path is given append the name of the private key to this path
    if path:
        private_key_path = os.path.join(path, name)
        with open(private_key_path, "wb") as f:
            f.write(pem)

    # generate public key and store it under the same name as .pub
    pub_name = name.split(".")[0] + ".pub"
    if path:
        pub_key_path = os.path.join(path, pub_name)
    else:
        pub_key_path = os.getcwd().join(pub_name)
    with open(pub_key_path, "wb") as f:
        f.write(
            private_key.public_key().public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )

    results = [GeneratorResult(loc=("central", "private_key"), value=private_key_path)]

    if password:
        results.append(
            GeneratorResult(loc=("central", "private_key_password"), value=password)
        )

    return results


def generate_fernet_key() -> GeneratorResult:
    key = Fernet.generate_key()

    result = GeneratorResult(loc=("api", "fernet_key"), value=key.decode())
    return result



import os
from enum import Enum
from typing import Tuple

import yaml


class ConfigFiles(Enum):
    """
    Enum for config file names
    """

    STATION_CONFIG = "station_config.yml"
    STATION_CONFIG_SHORT = "config.yml"
    STATION = "station.yml"


def load_config(file_name) -> dict:
    """
    Loads a config file and returns a dict with the content
    """
    with open(file_name, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

        except FileNotFoundError:
            FileNotFoundError(f"Config file {file_name} not found")


def find_config(directory) -> Tuple[dict, str]:
    """
    Finds a config file in a directory and returns a dict with the content
    """
    for file_name in ConfigFiles:
        try:
            return (
                load_config(os.path.join(directory, file_name.value)),
                file_name.value,
            )
        except FileNotFoundError:
            pass
    raise FileNotFoundError(f"No config file found in {directory}")



from datetime import datetime

from sqlalchemy import Boolean, Column, DateTime, Integer, String

from station.app.db.base_class import Base


class Notification(Base):
    __tablename__ = "notifications"
    id = Column(Integer, primary_key=True, index=True)
    target_user = Column(String, default="all")
    topic = Column(String, default="trains")
    title = Column(String, nullable=True)
    message = Column(String)
    is_read = Column(Boolean, default=False)
    type = Column(String, default="info")
    created_at = Column(DateTime, default=datetime.now())



import uuid

from station.app.db.base import Base
from station.app.db.session import SessionLocal, engine
from station.app.models import docker_trains


# TODO use alembic
def setup_db(dev=False, reset=False):

    if reset:
        reset_db(dev=False)
    else:
        Base.metadata.create_all(bind=engine)
    if dev:
        seed_db_for_testing()


def reset_db(dev=False):
    Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine)

    if dev:
        seed_db_for_testing()


def seed_db_for_testing():
    session = SessionLocal()
    # create docker trains
    if not session.query(docker_trains.DockerTrain).all():

        dts = []
        for _ in range(3):
            dt = docker_trains.DockerTrain(
                train_id=str(uuid.uuid4()),
            )
            dts.append(dt)
        session.add_all(dts)
        session.commit()

        # Create states for the created trains
        states = []
        for dt in dts:
            state = docker_trains.DockerTrainState(train_id=dt.id, status="inactive")
            states.append(state)

        session.add_all(states)

        config = docker_trains.DockerTrainConfig(name="default")

        session.add(config)
        session.commit()

    session.close()


if __name__ == "__main__":
    # Base.metadata.drop_all(bind=engine)
    setup_db()



import importlib.machinery
import importlib.util
import os
import tempfile
from typing import Any

import requests
from conductor_lib.src.torch import LightningTrainModel
from sqlalchemy.orm import Session

from station.app.crud import federated_trains
from station.common.clients.minio import MinioClient


class ModelLoader:
    def __init__(self, station_url: str = None):
        self.station_url = station_url if station_url else os.getenv("STATION_API_URL")
        assert self.station_url
        self.minio_client = MinioClient()

    def get_model_from_station(self, model_id: str = None) -> LightningTrainModel:
        """
        Use the station api to get a model registered under the specified model_id and return the initialized model

        :param model_id: identifier of model registered at a station
        :return: initialized model instance stored in the db with model_id
        """

        if not model_id:
            model_id = self.model_id

        # Get the model from the station api
        url = self.station_url + f"/api/station/models/{model_id}"

        r = requests.get(url=url)

        r.raise_for_status()
        model = r.json()
        # initialize the model based on the received src code and name
        model = self._make_lightning_module(model["model_src"], model["model_name"])
        return model

    def load_train_model(self, db: Session, train_id: Any):
        db_train = federated_trains.get_by_train_id(db=db, train_id=train_id)

        model = db_train.model
        assert model

        lightning_module = self._make_lightning_module(
            model.model_src, module_name=model.model_name
        )
        return lightning_module

    def save_model_checkpoint(self):
        # TODO
        pass

    def load_torch_model_from_json(
        self, model_json: dict = None
    ) -> LightningTrainModel:
        """
        Parse the content relevant for initializing a model from the given dictionary containing a user submitted
        json file defining the module.
        Initialize the model based on the found definitions.

        :param model_json: submitted json object
        :return: initialized instance of a LightningTrainModel
        """
        lightning_model_name = model_json["objects"]["lightning_model"]
        lightning_model_src = model_json["src"]["lightning_model"]

        ligntning_module = self._make_lightning_module(
            lightning_model_src, lightning_model_name
        )
        return ligntning_module

    @staticmethod
    def _make_lightning_module(src: str, module_name: str) -> LightningTrainModel:
        """
        Creates a temporary python file based on the src code string submitted using a client library.
        The lightning train model specified by module_name is imported and initialized from the created module and
        returned as an in memory object.

        :param src: String containing the source definition of the model
        :param module_name: the name of the model to be imported from src
        :return: an initialized LightningTrainModel instance based on the submitted source code
        """

        # intialize a tempfile with the model source code
        tmp_mod = tempfile.NamedTemporaryFile("w+b", suffix=".py", delete=False)
        tmp_mod.write(src.encode())
        tmp_mod.read()

        # initialize the temp file as a python module
        spec = importlib.util.spec_from_file_location("lightning_model", tmp_mod.name)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)

        # Get the model by its name and initialize it
        model: LightningTrainModel = foo.__getattribute__(module_name)()

        # cleanup
        tmp_mod.close()
        os.unlink(tmp_mod.name)

        return model



from fastapi import APIRouter

from station.app.api.api_v1.endpoints import (
    airflow,
    datasets,
    docker_trains,
    fhir,
    local_trains,
    master_images,
    notifications,
    station,
    station_status,
)

api_router = APIRouter()

# Include the routers defined in the endpoints file in the main api

api_router.include_router(station.router, prefix="/station", tags=["Station"])
api_router.include_router(
    docker_trains.router, prefix="/trains/docker", tags=["Trains", "PHT 1.0"]
)
api_router.include_router(datasets.router, prefix="/datasets", tags=["Datasets"])
api_router.include_router(
    station_status.router, prefix="/station/status", tags=["Station"]
)
api_router.include_router(
    local_trains.router, prefix="/local-trains", tags=["Local Trains"]
)
api_router.include_router(airflow.router, prefix="/airflow", tags=["Airflow"])
api_router.include_router(fhir.router, prefix="/fhir/server", tags=["FHIR"])
api_router.include_router(
    notifications.router, prefix="/notifications", tags=["Notifications"]
)
api_router.include_router(
    master_images.router, prefix="/master-images", tags=["Master Images"]
)



from fastapi.testclient import TestClient

from station.app.api.dependencies import get_db
from station.app.main import app

from .test_db import override_get_db

app.dependency_overrides[get_db] = override_get_db

client = TestClient(app)


def test_notification_create():
    """
    Test the creation of a Notification
    """
    response = client.post(
        "/api/notifications", json={"topic": "Test Notification", "message": "testing"}
    )
    assert response.status_code == 201
    assert response.json()["topic"] == "Test Notification"
    assert response.json()["message"] == "testing"


def test_notification_create_fail():
    """
    Test the creation of a Notification --fail
    """
    response = client.post(
        "/api/notification/x", json={"topic": "Test Notification", "message": "testing"}
    )
    assert response.status_code == 404, response.text


def test_notification_get_by_id():
    """
    Test the retrieval of a Notification
    """
    response = client.get("/api/notifications/1")
    assert response.status_code == 200
    assert response.json()["topic"] == "Test Notification"
    assert response.json()["message"] == "testing"


def test_notification_get_by_id_fail():
    """
    Test the retrieval of a Notification --fail
    """
    response = client.get("/api/notification/x")
    assert response.status_code == 404, response.text


def test_notification_update():
    """
    Test the update of a Notification
    """
    response = client.put(
        "/api/notifications/1",
        json={
            "message": "testing",
            "topic": "Test Notification Updated",
            "is_read": True,
        },
    )
    assert response.status_code == 200
    assert response.json()["topic"] == "Test Notification Updated"
    assert response.json()["message"] == "testing"


def test_get_list_notifications():
    """
    Test the retrieval of a list of Notifications
    """
    response = client.post(
        "/api/notifications",
        json={"topic": "Test Notification 2", "message": "testing"},
    )
    assert response.status_code == 201
    assert response.json()["topic"] == "Test Notification 2"
    assert response.json()["message"] == "testing"
    response = client.get("/api/notifications")
    assert response.status_code == 200
    assert len(response.json()) >= 1


def test_notification_delete():
    """
    Test the deletion of a Notification
    """
    response = client.delete("/api/notifications/1")
    assert response.status_code == 202, response.text



import os
import pickle

import numpy as np
import torchvision
from PIL import Image as pimg
from tqdm import tqdm

MINIO_PATH = "../../../data"
CIFAR_PATH = "../../data/cifar10/cifar-10-batches-py"


def load_cifar_10(download=True):
    cifar_path = os.path.join(MINIO_PATH, "cifar10")
    batch_path = os.path.join(cifar_path, "cifar-10-batches-py")

    # Download the torchvision cifar 10 dataset
    if download:
        torchvision.datasets.CIFAR10(cifar_path, download=True)

    # Extract image data and labels from the batches
    extract_cifar_10_batches(batch_path, cifar_path)


def extract_cifar_10_batches(batch_path: str, out_path: str):
    # load metadata
    categories = pickle.load(open(os.path.join(batch_path, "batches.meta"), "rb"))[
        "label_names"
    ]
    # Extract the 5 data batches
    for i in range(5):

        # create directory for batch if it doesn't exist
        if not os.path.isdir(os.path.join(out_path, f"batch_{i + 1}")):
            os.mkdir(os.path.join(out_path, f"batch_{i + 1}"))
        # load batch file
        file = f"data_batch_{i + 1}"
        images, labels, filenames = load_cifar_pickle(batch_path, file)
        # save images as png files
        print(f"Extracting images for batch {i}")
        for j, image in enumerate(tqdm(images)):

            string_label = str(categories[labels[j]])
            # Create class directory if it does not exist
            class_dir = os.path.join(out_path, f"batch_{i + 1}", string_label)
            if not os.path.isdir(class_dir):
                os.mkdir(class_dir)
            outfile = (
                out_path + f"/batch_{i + 1}/" + string_label + "/" + str(filenames[j])
            )
            # Transpose image to fit regular image dimensions
            img_array = image.transpose(1, 2, 0)
            img = pimg.fromarray(img_array, mode="RGB")
            img.save(outfile)


def load_cifar_pickle(path, file):
    f = open(os.path.join(path, file), "rb")
    dict = pickle.load(f, encoding="bytes")
    images = dict[b"data"]
    images = np.reshape(images, (10000, 3, 32, 32))
    labels = np.array(dict[b"labels"])
    filenames = np.array((dict[b"filenames"])).astype("str")
    return images, labels, filenames


if __name__ == "__main__":
    load_cifar_10(download=False)



import click

import docker
from station.common.constants import Icons


def start_server(name: str):
    name = _get_name_if_not_exists(name)
    client = docker.from_env()
    click.echo(f"Starting server {name}...", nl=False)
    try:
        container = client.containers.get(name)
        container.start()
        click.echo(Icons.CHECKMARK.value)
    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error starting server:\n {e}")


def stop_server(name: str):
    name = _get_name_if_not_exists(name)
    client = docker.from_env()
    click.echo(f"Stopping server {name}...", nl=False)
    try:
        container = client.containers.get(name)
        container.stop()
        click.echo(Icons.CHECKMARK.value)
    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error stopping server:\n {e}")


def remove_server(name: str):
    name = _get_name_if_not_exists(name)
    client = docker.from_env()
    click.echo(f"Removing server {name}...", nl=False)
    try:
        container = client.containers.get(name)
        container.remove()
        click.echo(Icons.CHECKMARK.value)
    except Exception as e:
        click.echo(Icons.CROSS.value)
        click.echo(f"Error removing server:\n {e}")


def list_servers():
    client = docker.from_env()
    client.containers.list()
    # todo
    click.echo("coming soon...")


def _get_name_if_not_exists(name: str) -> str:
    if name:
        return name
    else:
        name = click.prompt("Please enter the name of the server")
    return name



import os
from collections import namedtuple

import pytest
import yaml
from dotenv import find_dotenv, load_dotenv

from station.common.clients.central.central_client import CentralApiClient
from station.common.constants import PHTDirectories
from station.ctl.install import certs, docker, templates
from station.ctl.install.command import _setup_auth_server
from station.ctl.install.fs import create_pht_dirs
from station.ctl.util import get_template_env


@pytest.fixture
def central_client():
    load_dotenv(find_dotenv())
    url = os.getenv("CENTRAL_API_URL")
    client_id = os.getenv("STATION_ROBOT_ID")
    client_secret = os.getenv("STATION_ROBOT_SECRET")
    return CentralApiClient(url, client_id, client_secret)


def test_ensure_directory_structure(tmp_path):
    p = tmp_path / "station"
    create_pht_dirs(p)
    for dir in PHTDirectories:
        assert p.joinpath(dir.value).exists()


def test_get_template_env():
    env = get_template_env()
    assert env


def test_setup_auth():
    ctx = {
        "obj": {
            "version": "latest",
        }
    }
    ctx = namedtuple("Struct", ctx.keys())(*ctx.values())
    _setup_auth_server(ctx)


def test_render_init_sql():
    sql = templates.render_init_sql("test_db_user")
    assert sql
    print(sql)
    lines = sql.splitlines()
    assert len(lines) == 4
    assert lines[-1].endswith("test_db_user;")
    assert lines[-2].endswith("test_db_user;")


def test_render_traefik_configs():
    traefik_config, router_config = templates.render_traefik_configs(
        http_port=80,
        https_port=443,
        domain="test.com",
        https_enabled=True,
        certs=[{"cert": "test", "key": "test"}],
    )
    assert traefik_config
    assert router_config

    print(traefik_config)
    print(router_config)

    traefik_dict = yaml.safe_load(traefik_config)
    router_dict = yaml.safe_load(router_config)
    print(traefik_dict)
    print(router_dict)

    assert traefik_dict["entryPoints"]["http"]["address"] == ":80"
    assert traefik_dict["entryPoints"]["https"]["address"] == ":443"

    assert (
        router_dict["http"]["routers"]["traefik"]["tls"]["domains"][0]["main"]
        == "test.com"
    )
    assert router_dict["tls"]["certificates"][0]["certFile"] == "test"
    assert router_dict["tls"]["certificates"][0]["keyFile"] == "test"


def test_setup_volumes():
    docker.setup_docker()


def test_generate_certs(tmp_path):
    key_path = tmp_path / "key.pem"
    cert_path = tmp_path / "cert.pem"
    certs.generate_certificates("test.com", str(key_path), str(cert_path))
    assert key_path.exists()
    assert key_path.read_bytes()
    assert cert_path.exists()
    assert cert_path.read_bytes()



import tarfile
import time
from io import BytesIO

import pytest

from station.app.trains.local.build import build_train
from station.app.trains.local.docker import make_docker_file


@pytest.fixture
def simple_file_archive() -> BytesIO:
    archive = BytesIO()
    with tarfile.TarFile(fileobj=archive, mode="w") as tar:
        content = "import os\n\nif __name__ == '__main__':\n    print('hello')\n"
        file = BytesIO(content.encode("utf-8"))
        info = tarfile.TarInfo(name="entrypoint.py")
        info.size = len(file.getbuffer())
        info.mtime = time.time()
        tar.addfile(info, file)

    archive.seek(0)
    return archive


def test_make_docker_file():
    master_image = "ubuntu:latest"
    entrypoint_file = "entrypoint.sh"
    command = "bash"
    command_args = ["-c", "echo 'hello'"]

    docker_file = make_docker_file(master_image, entrypoint_file, command, command_args)

    print(docker_file.read())

    docker_file = make_docker_file(master_image, entrypoint_file, command)
    print(docker_file.read())


def test_build_basic(simple_file_archive):

    train = build_train(
        train_id="test",
        master_image_id="python:3.8",
        files=simple_file_archive,
        command="python",
        entrypoint_file="entrypoint.py",
    )

    print(train)


def test_api_create():
    pass
